/*
 *  db.h
 *
 *  Copyright (c) 2006-2019 Pacman Development Team <pacman-dev@archlinux.org>
 *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
 *  Copyright (c) 2005 by Aurelien Foret <orelien@chez.com>
 *  Copyright (c) 2006 by Miklos Vajna <vmiklos@frugalware.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef HBPM_DB_H
#define HBPM_DB_H

/* libarchive */
#include <archive.h>
#include <archive_entry.h>

#include "hbpm.h"
#include "pkghash.h"
#include "signing.h"

/* Database entries */
typedef enum _hbpm_dbinfrq_t {
	INFRQ_BASE = (1 << 0),
	INFRQ_DESC = (1 << 1),
	INFRQ_FILES = (1 << 2),
	INFRQ_SCRIPTLET = (1 << 3),
	INFRQ_DSIZE = (1 << 4),
	/* ALL should be info stored in the package or database */
	INFRQ_ALL = INFRQ_BASE | INFRQ_DESC | INFRQ_FILES |
		INFRQ_SCRIPTLET | INFRQ_DSIZE,
	INFRQ_ERROR = (1 << 30)
} hbpm_dbinfrq_t;

/** Database status. Bitflags. */
enum _hbpm_dbstatus_t {
	DB_STATUS_VALID = (1 << 0),
	DB_STATUS_INVALID = (1 << 1),
	DB_STATUS_EXISTS = (1 << 2),
	DB_STATUS_MISSING = (1 << 3),

	DB_STATUS_LOCAL = (1 << 10),
	DB_STATUS_PKGCACHE = (1 << 11),
	DB_STATUS_GRPCACHE = (1 << 12)
};

struct db_operations {
	int (*validate) (hbpm_db_t *);
	int (*populate) (hbpm_db_t *);
	void (*unregister) (hbpm_db_t *);
};

/* Database */
struct __hbpm_db_t {
	hbpm_handle_t *handle;
	char *treename;
	/* do not access directly, use _hbpm_db_path(db) for lazy access */
	char *_path;
	hbpm_pkghash_t *pkgcache;
	hbpm_list_t *grpcache;
	hbpm_list_t *servers;
	struct db_operations *ops;

	/* bitfields for validity, local, loaded caches, etc. */
	/* From _hbpm_dbstatus_t */
	int status;
	/* hbpm_siglevel_t */
	int siglevel;
	/* hbpm_db_usage_t */
	int usage;
};


/* db.c, database general calls */
hbpm_db_t *_hbpm_db_new(const char *treename, int is_local);
void _hbpm_db_free(hbpm_db_t *db);
const char *_hbpm_db_path(hbpm_db_t *db);
int _hbpm_db_cmp(const void *d1, const void *d2);
int _hbpm_db_search(hbpm_db_t *db, const hbpm_list_t *needles,
		hbpm_list_t **ret);
hbpm_db_t *_hbpm_db_register_local(hbpm_handle_t *handle);
hbpm_db_t *_hbpm_db_register_sync(hbpm_handle_t *handle, const char *treename,
		int level);
void _hbpm_db_unregister(hbpm_db_t *db);

/* be_*.c, backend specific calls */
int _hbpm_local_db_prepare(hbpm_db_t *db, hbpm_pkg_t *info);
int _hbpm_local_db_write(hbpm_db_t *db, hbpm_pkg_t *info, int inforeq);
int _hbpm_local_db_remove(hbpm_db_t *db, hbpm_pkg_t *info);
char *_hbpm_local_db_pkgpath(hbpm_db_t *db, hbpm_pkg_t *info, const char *filename);

/* cache bullshit */
/* packages */
void _hbpm_db_free_pkgcache(hbpm_db_t *db);
int _hbpm_db_add_pkgincache(hbpm_db_t *db, hbpm_pkg_t *pkg);
int _hbpm_db_remove_pkgfromcache(hbpm_db_t *db, hbpm_pkg_t *pkg);
hbpm_pkghash_t *_hbpm_db_get_pkgcache_hash(hbpm_db_t *db);
hbpm_list_t *_hbpm_db_get_pkgcache(hbpm_db_t *db);
hbpm_pkg_t *_hbpm_db_get_pkgfromcache(hbpm_db_t *db, const char *target);
/* groups */
hbpm_list_t *_hbpm_db_get_groupcache(hbpm_db_t *db);
hbpm_group_t *_hbpm_db_get_groupfromcache(hbpm_db_t *db, const char *target);

#endif /* HBPM_DB_H */
