/*
 *  handle.c
 *
 *  Copyright (c) 2006-2019 Pacman Development Team <pacman-dev@archlinux.org>
 *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
 *  Copyright (c) 2005 by Aurelien Foret <orelien@chez.com>
 *  Copyright (c) 2005, 2006 by Miklos Vajna <vmiklos@frugalware.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <syslog.h>
#include <sys/stat.h>
#include <fcntl.h>

/* libhbpm */
#include "handle.h"
#include "hbpm_list.h"
#include "util.h"
#include "log.h"
#include "trans.h"
#include "hbpm.h"
#include "deps.h"

hbpm_handle_t *_hbpm_handle_new(void)
{
	hbpm_handle_t *handle;

	CALLOC(handle, 1, sizeof(hbpm_handle_t), return NULL);
	handle->lockfd = -1;

	return handle;
}

void _hbpm_handle_free(hbpm_handle_t *handle)
{
	if(handle == NULL) {
		return;
	}

	/* close logfile */
	if(handle->logstream) {
		fclose(handle->logstream);
		handle->logstream = NULL;
	}
	if(handle->usesyslog) {
		handle->usesyslog = 0;
		closelog();
	}

#ifdef HAVE_LIBCURL
	/* release curl handle */
	curl_easy_cleanup(handle->curl);
#endif

#ifdef HAVE_LIBGPGME
	FREELIST(handle->known_keys);
#endif

	/* free memory */
	_hbpm_trans_free(handle->trans);
	FREE(handle->root);
	FREE(handle->dbpath);
	FREE(handle->dbext);
	FREELIST(handle->cachedirs);
	FREELIST(handle->hookdirs);
	FREE(handle->logfile);
	FREE(handle->lockfile);
	FREE(handle->arch);
	FREE(handle->gpgdir);
	FREELIST(handle->noupgrade);
	FREELIST(handle->noextract);
	FREELIST(handle->ignorepkg);
	FREELIST(handle->ignoregroup);
	FREELIST(handle->overwrite_files);

	hbpm_list_free_inner(handle->assumeinstalled, (hbpm_list_fn_free)hbpm_dep_free);
	hbpm_list_free(handle->assumeinstalled);

	FREE(handle);
}

/** Lock the database */
int _hbpm_handle_lock(hbpm_handle_t *handle)
{
	char *dir, *ptr;

	ASSERT(handle->lockfile != NULL, return -1);
	ASSERT(handle->lockfd < 0, return 0);

	/* create the dir of the lockfile first */
	STRDUP(dir, handle->lockfile, return -1);
	ptr = strrchr(dir, '/');
	if(ptr) {
		*ptr = '\0';
	}
	if(_hbpm_makepath(dir)) {
		FREE(dir);
		return -1;
	}
	FREE(dir);

	do {
		handle->lockfd = open(handle->lockfile, O_WRONLY | O_CREAT | O_EXCL | O_CLOEXEC, 0000);
	} while(handle->lockfd == -1 && errno == EINTR);

	return (handle->lockfd >= 0 ? 0 : -1);
}

int SYMEXPORT hbpm_unlock(hbpm_handle_t *handle)
{
	ASSERT(handle != NULL, return -1);
	ASSERT(handle->lockfile != NULL, return 0);
	ASSERT(handle->lockfd >= 0, return 0);

	close(handle->lockfd);
	handle->lockfd = -1;

	if(unlink(handle->lockfile) != 0) {
		RET_ERR_ASYNC_SAFE(handle, HBPM_ERR_SYSTEM, -1);
	} else {
		return 0;
	}
}

int _hbpm_handle_unlock(hbpm_handle_t *handle)
{
	if(hbpm_unlock(handle) != 0) {
		if(errno == ENOENT) {
			_hbpm_log(handle, HBPM_LOG_WARNING,
					_("lock file missing %s\n"), handle->lockfile);
			hbpm_logaction(handle, HBPM_CALLER_PREFIX,
					"warning: lock file missing %s\n", handle->lockfile);
			return 0;
		} else {
			_hbpm_log(handle, HBPM_LOG_WARNING,
					_("could not remove lock file %s\n"), handle->lockfile);
			hbpm_logaction(handle, HBPM_CALLER_PREFIX,
					"warning: could not remove lock file %s\n", handle->lockfile);
			return -1;
		}
	}

	return 0;
}


hbpm_cb_log SYMEXPORT hbpm_option_get_logcb(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->logcb;
}

hbpm_cb_download SYMEXPORT hbpm_option_get_dlcb(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->dlcb;
}

hbpm_cb_fetch SYMEXPORT hbpm_option_get_fetchcb(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->fetchcb;
}

hbpm_cb_totaldl SYMEXPORT hbpm_option_get_totaldlcb(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->totaldlcb;
}

hbpm_cb_event SYMEXPORT hbpm_option_get_eventcb(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->eventcb;
}

hbpm_cb_question SYMEXPORT hbpm_option_get_questioncb(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->questioncb;
}

hbpm_cb_progress SYMEXPORT hbpm_option_get_progresscb(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->progresscb;
}

const char SYMEXPORT *hbpm_option_get_root(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->root;
}

const char SYMEXPORT *hbpm_option_get_dbpath(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->dbpath;
}

hbpm_list_t SYMEXPORT *hbpm_option_get_hookdirs(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->hookdirs;
}

hbpm_list_t SYMEXPORT *hbpm_option_get_cachedirs(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->cachedirs;
}

const char SYMEXPORT *hbpm_option_get_logfile(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->logfile;
}

const char SYMEXPORT *hbpm_option_get_lockfile(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->lockfile;
}

const char SYMEXPORT *hbpm_option_get_gpgdir(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->gpgdir;
}

int SYMEXPORT hbpm_option_get_usesyslog(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return -1);
	return handle->usesyslog;
}

hbpm_list_t SYMEXPORT *hbpm_option_get_noupgrades(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->noupgrade;
}

hbpm_list_t SYMEXPORT *hbpm_option_get_noextracts(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->noextract;
}

hbpm_list_t SYMEXPORT *hbpm_option_get_ignorepkgs(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->ignorepkg;
}

hbpm_list_t SYMEXPORT *hbpm_option_get_ignoregroups(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->ignoregroup;
}

hbpm_list_t SYMEXPORT *hbpm_option_get_overwrite_files(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->overwrite_files;
}

hbpm_list_t SYMEXPORT *hbpm_option_get_assumeinstalled(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->assumeinstalled;
}

const char SYMEXPORT *hbpm_option_get_arch(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->arch;
}

int SYMEXPORT hbpm_option_get_checkspace(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return -1);
	return handle->checkspace;
}

const char SYMEXPORT *hbpm_option_get_dbext(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->dbext;
}

int SYMEXPORT hbpm_option_set_logcb(hbpm_handle_t *handle, hbpm_cb_log cb)
{
	CHECK_HANDLE(handle, return -1);
	handle->logcb = cb;
	return 0;
}

int SYMEXPORT hbpm_option_set_dlcb(hbpm_handle_t *handle, hbpm_cb_download cb)
{
	CHECK_HANDLE(handle, return -1);
	handle->dlcb = cb;
	return 0;
}

int SYMEXPORT hbpm_option_set_fetchcb(hbpm_handle_t *handle, hbpm_cb_fetch cb)
{
	CHECK_HANDLE(handle, return -1);
	handle->fetchcb = cb;
	return 0;
}

int SYMEXPORT hbpm_option_set_totaldlcb(hbpm_handle_t *handle, hbpm_cb_totaldl cb)
{
	CHECK_HANDLE(handle, return -1);
	handle->totaldlcb = cb;
	return 0;
}

int SYMEXPORT hbpm_option_set_eventcb(hbpm_handle_t *handle, hbpm_cb_event cb)
{
	CHECK_HANDLE(handle, return -1);
	handle->eventcb = cb;
	return 0;
}

int SYMEXPORT hbpm_option_set_questioncb(hbpm_handle_t *handle, hbpm_cb_question cb)
{
	CHECK_HANDLE(handle, return -1);
	handle->questioncb = cb;
	return 0;
}

int SYMEXPORT hbpm_option_set_progresscb(hbpm_handle_t *handle, hbpm_cb_progress cb)
{
	CHECK_HANDLE(handle, return -1);
	handle->progresscb = cb;
	return 0;
}

static char *canonicalize_path(const char *path)
{
	char *new_path;
	size_t len;

	/* verify path ends in a '/' */
	len = strlen(path);
	if(path[len - 1] != '/') {
		len += 1;
	}
	CALLOC(new_path, len + 1, sizeof(char), return NULL);
	strcpy(new_path, path);
	new_path[len - 1] = '/';
	return new_path;
}

hbpm_errno_t _hbpm_set_directory_option(const char *value,
		char **storage, int must_exist)
{
	struct stat st;
	char real[PATH_MAX];
	const char *path;

	path = value;
	if(!path) {
		return HBPM_ERR_WRONG_ARGS;
	}
	if(must_exist) {
		if(stat(path, &st) == -1 || !S_ISDIR(st.st_mode)) {
			return HBPM_ERR_NOT_A_DIR;
		}
		if(!realpath(path, real)) {
			return HBPM_ERR_NOT_A_DIR;
		}
		path = real;
	}

	if(*storage) {
		FREE(*storage);
	}
	*storage = canonicalize_path(path);
	if(!*storage) {
		return HBPM_ERR_MEMORY;
	}
	return 0;
}

int SYMEXPORT hbpm_option_add_hookdir(hbpm_handle_t *handle, const char *hookdir)
{
	char *newhookdir;

	CHECK_HANDLE(handle, return -1);
	ASSERT(hookdir != NULL, RET_ERR(handle, HBPM_ERR_WRONG_ARGS, -1));

	newhookdir = canonicalize_path(hookdir);
	if(!newhookdir) {
		RET_ERR(handle, HBPM_ERR_MEMORY, -1);
	}
	handle->hookdirs = hbpm_list_add(handle->hookdirs, newhookdir);
	_hbpm_log(handle, HBPM_LOG_DEBUG, "option 'hookdir' = %s\n", newhookdir);
	return 0;
}

int SYMEXPORT hbpm_option_set_hookdirs(hbpm_handle_t *handle, hbpm_list_t *hookdirs)
{
	hbpm_list_t *i;
	CHECK_HANDLE(handle, return -1);
	if(handle->hookdirs) {
		FREELIST(handle->hookdirs);
	}
	for(i = hookdirs; i; i = i->next) {
		int ret = hbpm_option_add_hookdir(handle, i->data);
		if(ret) {
			return ret;
		}
	}
	return 0;
}

int SYMEXPORT hbpm_option_remove_hookdir(hbpm_handle_t *handle, const char *hookdir)
{
	char *vdata = NULL;
	char *newhookdir;
	CHECK_HANDLE(handle, return -1);
	ASSERT(hookdir != NULL, RET_ERR(handle, HBPM_ERR_WRONG_ARGS, -1));

	newhookdir = canonicalize_path(hookdir);
	if(!newhookdir) {
		RET_ERR(handle, HBPM_ERR_MEMORY, -1);
	}
	handle->hookdirs = hbpm_list_remove_str(handle->hookdirs, newhookdir, &vdata);
	FREE(newhookdir);
	if(vdata != NULL) {
		FREE(vdata);
		return 1;
	}
	return 0;
}

int SYMEXPORT hbpm_option_add_cachedir(hbpm_handle_t *handle, const char *cachedir)
{
	char *newcachedir;

	CHECK_HANDLE(handle, return -1);
	ASSERT(cachedir != NULL, RET_ERR(handle, HBPM_ERR_WRONG_ARGS, -1));
	/* don't stat the cachedir yet, as it may not even be needed. we can
	 * fail later if it is needed and the path is invalid. */

	newcachedir = canonicalize_path(cachedir);
	if(!newcachedir) {
		RET_ERR(handle, HBPM_ERR_MEMORY, -1);
	}
	handle->cachedirs = hbpm_list_add(handle->cachedirs, newcachedir);
	_hbpm_log(handle, HBPM_LOG_DEBUG, "option 'cachedir' = %s\n", newcachedir);
	return 0;
}

int SYMEXPORT hbpm_option_set_cachedirs(hbpm_handle_t *handle, hbpm_list_t *cachedirs)
{
	hbpm_list_t *i;
	CHECK_HANDLE(handle, return -1);
	if(handle->cachedirs) {
		FREELIST(handle->cachedirs);
	}
	for(i = cachedirs; i; i = i->next) {
		int ret = hbpm_option_add_cachedir(handle, i->data);
		if(ret) {
			return ret;
		}
	}
	return 0;
}

int SYMEXPORT hbpm_option_remove_cachedir(hbpm_handle_t *handle, const char *cachedir)
{
	char *vdata = NULL;
	char *newcachedir;
	CHECK_HANDLE(handle, return -1);
	ASSERT(cachedir != NULL, RET_ERR(handle, HBPM_ERR_WRONG_ARGS, -1));

	newcachedir = canonicalize_path(cachedir);
	if(!newcachedir) {
		RET_ERR(handle, HBPM_ERR_MEMORY, -1);
	}
	handle->cachedirs = hbpm_list_remove_str(handle->cachedirs, newcachedir, &vdata);
	FREE(newcachedir);
	if(vdata != NULL) {
		FREE(vdata);
		return 1;
	}
	return 0;
}

int SYMEXPORT hbpm_option_set_logfile(hbpm_handle_t *handle, const char *logfile)
{
	char *oldlogfile = handle->logfile;

	CHECK_HANDLE(handle, return -1);
	if(!logfile) {
		handle->pm_errno = HBPM_ERR_WRONG_ARGS;
		return -1;
	}

	STRDUP(handle->logfile, logfile, RET_ERR(handle, HBPM_ERR_MEMORY, -1));

	/* free the old logfile path string, and close the stream so logaction
	 * will reopen a new stream on the new logfile */
	if(oldlogfile) {
		FREE(oldlogfile);
	}
	if(handle->logstream) {
		fclose(handle->logstream);
		handle->logstream = NULL;
	}
	_hbpm_log(handle, HBPM_LOG_DEBUG, "option 'logfile' = %s\n", handle->logfile);
	return 0;
}

int SYMEXPORT hbpm_option_set_gpgdir(hbpm_handle_t *handle, const char *gpgdir)
{
	int err;
	CHECK_HANDLE(handle, return -1);
	if((err = _hbpm_set_directory_option(gpgdir, &(handle->gpgdir), 0))) {
		RET_ERR(handle, err, -1);
	}
	_hbpm_log(handle, HBPM_LOG_DEBUG, "option 'gpgdir' = %s\n", handle->gpgdir);
	return 0;
}

int SYMEXPORT hbpm_option_set_usesyslog(hbpm_handle_t *handle, int usesyslog)
{
	CHECK_HANDLE(handle, return -1);
	handle->usesyslog = usesyslog;
	return 0;
}

static int _hbpm_option_strlist_add(hbpm_handle_t *handle, hbpm_list_t **list, const char *str)
{
	char *dup;
	CHECK_HANDLE(handle, return -1);
	STRDUP(dup, str, RET_ERR(handle, HBPM_ERR_MEMORY, -1));
	*list = hbpm_list_add(*list, dup);
	return 0;
}

static int _hbpm_option_strlist_set(hbpm_handle_t *handle, hbpm_list_t **list, hbpm_list_t *newlist)
{
	CHECK_HANDLE(handle, return -1);
	FREELIST(*list);
	*list = hbpm_list_strdup(newlist);
	return 0;
}

static int _hbpm_option_strlist_rem(hbpm_handle_t *handle, hbpm_list_t **list, const char *str)
{
	char *vdata = NULL;
	CHECK_HANDLE(handle, return -1);
	*list = hbpm_list_remove_str(*list, str, &vdata);
	if(vdata != NULL) {
		FREE(vdata);
		return 1;
	}
	return 0;
}

int SYMEXPORT hbpm_option_add_noupgrade(hbpm_handle_t *handle, const char *pkg)
{
	return _hbpm_option_strlist_add(handle, &(handle->noupgrade), pkg);
}

int SYMEXPORT hbpm_option_set_noupgrades(hbpm_handle_t *handle, hbpm_list_t *noupgrade)
{
	return _hbpm_option_strlist_set(handle, &(handle->noupgrade), noupgrade);
}

int SYMEXPORT hbpm_option_remove_noupgrade(hbpm_handle_t *handle, const char *pkg)
{
	return _hbpm_option_strlist_rem(handle, &(handle->noupgrade), pkg);
}

int SYMEXPORT hbpm_option_match_noupgrade(hbpm_handle_t *handle, const char *path)
{
	return _hbpm_fnmatch_patterns(handle->noupgrade, path);
}

int SYMEXPORT hbpm_option_add_noextract(hbpm_handle_t *handle, const char *path)
{
	return _hbpm_option_strlist_add(handle, &(handle->noextract), path);
}

int SYMEXPORT hbpm_option_set_noextracts(hbpm_handle_t *handle, hbpm_list_t *noextract)
{
	return _hbpm_option_strlist_set(handle, &(handle->noextract), noextract);
}

int SYMEXPORT hbpm_option_remove_noextract(hbpm_handle_t *handle, const char *path)
{
	return _hbpm_option_strlist_rem(handle, &(handle->noextract), path);
}

int SYMEXPORT hbpm_option_match_noextract(hbpm_handle_t *handle, const char *path)
{
	return _hbpm_fnmatch_patterns(handle->noextract, path);
}

int SYMEXPORT hbpm_option_add_ignorepkg(hbpm_handle_t *handle, const char *pkg)
{
	return _hbpm_option_strlist_add(handle, &(handle->ignorepkg), pkg);
}

int SYMEXPORT hbpm_option_set_ignorepkgs(hbpm_handle_t *handle, hbpm_list_t *ignorepkgs)
{
	return _hbpm_option_strlist_set(handle, &(handle->ignorepkg), ignorepkgs);
}

int SYMEXPORT hbpm_option_remove_ignorepkg(hbpm_handle_t *handle, const char *pkg)
{
	return _hbpm_option_strlist_rem(handle, &(handle->ignorepkg), pkg);
}

int SYMEXPORT hbpm_option_add_ignoregroup(hbpm_handle_t *handle, const char *grp)
{
	return _hbpm_option_strlist_add(handle, &(handle->ignoregroup), grp);
}

int SYMEXPORT hbpm_option_set_ignoregroups(hbpm_handle_t *handle, hbpm_list_t *ignoregrps)
{
	return _hbpm_option_strlist_set(handle, &(handle->ignoregroup), ignoregrps);
}

int SYMEXPORT hbpm_option_remove_ignoregroup(hbpm_handle_t *handle, const char *grp)
{
	return _hbpm_option_strlist_rem(handle, &(handle->ignoregroup), grp);
}

int SYMEXPORT hbpm_option_add_overwrite_file(hbpm_handle_t *handle, const char *glob)
{
	return _hbpm_option_strlist_add(handle, &(handle->overwrite_files), glob);
}

int SYMEXPORT hbpm_option_set_overwrite_files(hbpm_handle_t *handle, hbpm_list_t *globs)
{
	return _hbpm_option_strlist_set(handle, &(handle->overwrite_files), globs);
}

int SYMEXPORT hbpm_option_remove_overwrite_file(hbpm_handle_t *handle, const char *glob)
{
	return _hbpm_option_strlist_rem(handle, &(handle->overwrite_files), glob);
}

int SYMEXPORT hbpm_option_add_assumeinstalled(hbpm_handle_t *handle, const hbpm_depend_t *dep)
{
	hbpm_depend_t *depcpy;
	CHECK_HANDLE(handle, return -1);
	ASSERT(dep->mod == HBPM_DEP_MOD_EQ || dep->mod == HBPM_DEP_MOD_ANY,
			RET_ERR(handle, HBPM_ERR_WRONG_ARGS, -1));
	ASSERT((depcpy = _hbpm_dep_dup(dep)), RET_ERR(handle, HBPM_ERR_MEMORY, -1));

	/* fill in name_hash in case dep was built by hand */
	depcpy->name_hash = _hbpm_hash_sdbm(dep->name);
	handle->assumeinstalled = hbpm_list_add(handle->assumeinstalled, depcpy);
	return 0;
}

int SYMEXPORT hbpm_option_set_assumeinstalled(hbpm_handle_t *handle, hbpm_list_t *deps)
{
	CHECK_HANDLE(handle, return -1);
	if(handle->assumeinstalled) {
		hbpm_list_free_inner(handle->assumeinstalled, (hbpm_list_fn_free)hbpm_dep_free);
		hbpm_list_free(handle->assumeinstalled);
		handle->assumeinstalled = NULL;
	}
	while(deps) {
		if(hbpm_option_add_assumeinstalled(handle, deps->data) != 0) {
			return -1;
		}
		deps = deps->next;
	}
	return 0;
}

static int assumeinstalled_cmp(const void *d1, const void *d2)
{
	const hbpm_depend_t *dep1 = d1;
	const hbpm_depend_t *dep2 = d2;

	if(dep1->name_hash != dep2->name_hash
			|| strcmp(dep1->name, dep2->name) != 0) {
		return -1;
	}

	if(dep1->version && dep2->version
			&& strcmp(dep1->version, dep2->version) == 0) {
		return 0;
	}

	if(dep1->version == NULL && dep2->version == NULL) {
		return 0;
	}


	return -1;
}

int SYMEXPORT hbpm_option_remove_assumeinstalled(hbpm_handle_t *handle, const hbpm_depend_t *dep)
{
	hbpm_depend_t *vdata = NULL;
	CHECK_HANDLE(handle, return -1);

	handle->assumeinstalled = hbpm_list_remove(handle->assumeinstalled, dep, &assumeinstalled_cmp, (void **)&vdata);
	if(vdata != NULL) {
		hbpm_dep_free(vdata);
		return 1;
	}

	return 0;
}

int SYMEXPORT hbpm_option_set_arch(hbpm_handle_t *handle, const char *arch)
{
	CHECK_HANDLE(handle, return -1);
	if(handle->arch) FREE(handle->arch);
	STRDUP(handle->arch, arch, RET_ERR(handle, HBPM_ERR_MEMORY, -1));
	return 0;
}

hbpm_db_t SYMEXPORT *hbpm_get_localdb(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->db_local;
}

hbpm_list_t SYMEXPORT *hbpm_get_syncdbs(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return NULL);
	return handle->dbs_sync;
}

int SYMEXPORT hbpm_option_set_checkspace(hbpm_handle_t *handle, int checkspace)
{
	CHECK_HANDLE(handle, return -1);
	handle->checkspace = checkspace;
	return 0;
}

int SYMEXPORT hbpm_option_set_dbext(hbpm_handle_t *handle, const char *dbext)
{
	CHECK_HANDLE(handle, return -1);
	ASSERT(dbext, RET_ERR(handle, HBPM_ERR_WRONG_ARGS, -1));

	if(handle->dbext) {
		FREE(handle->dbext);
	}

	STRDUP(handle->dbext, dbext, RET_ERR(handle, HBPM_ERR_MEMORY, -1));

	_hbpm_log(handle, HBPM_LOG_DEBUG, "option 'dbext' = %s\n", handle->dbext);
	return 0;
}

int SYMEXPORT hbpm_option_set_default_siglevel(hbpm_handle_t *handle,
		int level)
{
	CHECK_HANDLE(handle, return -1);
	if(level == HBPM_SIG_USE_DEFAULT) {
		RET_ERR(handle, HBPM_ERR_WRONG_ARGS, -1);
	}
#ifdef HAVE_LIBGPGME
	handle->siglevel = level;
#else
	if(level != 0) {
		RET_ERR(handle, HBPM_ERR_MISSING_CAPABILITY_SIGNATURES, -1);
	}
#endif
	return 0;
}

int SYMEXPORT hbpm_option_get_default_siglevel(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return -1);
	return handle->siglevel;
}

int SYMEXPORT hbpm_option_set_local_file_siglevel(hbpm_handle_t *handle,
		int level)
{
	CHECK_HANDLE(handle, return -1);
#ifdef HAVE_LIBGPGME
	handle->localfilesiglevel = level;
#else
	if(level != 0 && level != HBPM_SIG_USE_DEFAULT) {
		RET_ERR(handle, HBPM_ERR_MISSING_CAPABILITY_SIGNATURES, -1);
	}
#endif
	return 0;
}

int SYMEXPORT hbpm_option_get_local_file_siglevel(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return -1);
	if(handle->localfilesiglevel & HBPM_SIG_USE_DEFAULT) {
		return handle->siglevel;
	} else {
		return handle->localfilesiglevel;
	}
}

int SYMEXPORT hbpm_option_set_remote_file_siglevel(hbpm_handle_t *handle,
		int level)
{
	CHECK_HANDLE(handle, return -1);
#ifdef HAVE_LIBGPGME
	handle->remotefilesiglevel = level;
#else
	if(level != 0 && level != HBPM_SIG_USE_DEFAULT) {
		RET_ERR(handle, HBPM_ERR_MISSING_CAPABILITY_SIGNATURES, -1);
	}
#endif
	return 0;
}

int SYMEXPORT hbpm_option_get_remote_file_siglevel(hbpm_handle_t *handle)
{
	CHECK_HANDLE(handle, return -1);
	if(handle->remotefilesiglevel & HBPM_SIG_USE_DEFAULT) {
		return handle->siglevel;
	} else {
		return handle->remotefilesiglevel;
	}
}

int SYMEXPORT hbpm_option_set_disable_dl_timeout(hbpm_handle_t *handle,
		unsigned short disable_dl_timeout)
{
	CHECK_HANDLE(handle, return -1);
#ifdef HAVE_LIBCURL
	handle->disable_dl_timeout = disable_dl_timeout;
#else
	(void)disable_dl_timeout; /* silence unused variable warnings */
#endif
	return 0;
}
