/*
 *  handle.h
 *
 *  Copyright (c) 2006-2019 Pacman Development Team <pacman-dev@archlinux.org>
 *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef HBPM_HANDLE_H
#define HBPM_HANDLE_H

#include <stdio.h>
#include <sys/types.h>
#include <regex.h>

#include "hbpm_list.h"
#include "hbpm.h"

#ifdef HAVE_LIBCURL
#include <curl/curl.h>
#endif

#define EVENT(h, e) \
do { \
	if((h)->eventcb) { \
		(h)->eventcb((hbpm_event_t *) (e)); \
	} \
} while(0)
#define QUESTION(h, q) \
do { \
	if((h)->questioncb) { \
		(h)->questioncb((hbpm_question_t *) (q)); \
	} \
} while(0)
#define PROGRESS(h, e, p, per, n, r) \
do { \
	if((h)->progresscb) { \
		(h)->progresscb(e, p, per, n, r); \
	} \
} while(0)

struct __hbpm_handle_t {
	/* internal usage */
	hbpm_db_t *db_local;    /* local db pointer */
	hbpm_list_t *dbs_sync;  /* List of (hbpm_db_t *) */
	FILE *logstream;        /* log file stream pointer */
	hbpm_trans_t *trans;

#ifdef HAVE_LIBCURL
	/* libcurl handle */
	CURL *curl;             /* reusable curl_easy handle */
	unsigned short disable_dl_timeout;
#endif

#ifdef HAVE_LIBGPGME
	hbpm_list_t *known_keys;  /* keys verified to be in our keychain */
#endif

	/* callback functions */
	hbpm_cb_log logcb;          /* Log callback function */
	hbpm_cb_download dlcb;      /* Download callback function */
	hbpm_cb_totaldl totaldlcb;  /* Total download callback function */
	hbpm_cb_fetch fetchcb;      /* Download file callback function */
	hbpm_cb_event eventcb;
	hbpm_cb_question questioncb;
	hbpm_cb_progress progresscb;

	/* filesystem paths */
	char *root;              /* Root path, default '/' */
	char *dbpath;            /* Base path to hyperman's DBs */
	char *logfile;           /* Name of the log file */
	char *lockfile;          /* Name of the lock file */
	char *gpgdir;            /* Directory where GnuPG files are stored */
	hbpm_list_t *cachedirs;  /* Paths to hyperman cache directories */
	hbpm_list_t *hookdirs;   /* Paths to hook directories */
	hbpm_list_t *overwrite_files; /* Paths that may be overwritten */

	/* package lists */
	hbpm_list_t *noupgrade;   /* List of packages NOT to be upgraded */
	hbpm_list_t *noextract;   /* List of files NOT to extract */
	hbpm_list_t *ignorepkg;   /* List of packages to ignore */
	hbpm_list_t *ignoregroup; /* List of groups to ignore */
	hbpm_list_t *assumeinstalled;   /* List of virtual packages used to satisfy dependencies */

	/* options */
	char *arch;              /* Architecture of packages we should allow */
	int usesyslog;           /* Use syslog instead of logfile? */ /* TODO move to frontend */
	int checkspace;          /* Check disk space before installing */
	char *dbext;             /* Sync DB extension */
	int siglevel;            /* Default signature verification level */
	int localfilesiglevel;   /* Signature verification level for local file
	                                       upgrade operations */
	int remotefilesiglevel;  /* Signature verification level for remote file
	                                       upgrade operations */

	/* error code */
	hbpm_errno_t pm_errno;

	/* lock file descriptor */
	int lockfd;
};

hbpm_handle_t *_hbpm_handle_new(void);
void _hbpm_handle_free(hbpm_handle_t *handle);

int _hbpm_handle_lock(hbpm_handle_t *handle);
int _hbpm_handle_unlock(hbpm_handle_t *handle);

hbpm_errno_t _hbpm_set_directory_option(const char *value,
		char **storage, int must_exist);

#endif /* HBPM_HANDLE_H */
