/*
 * hbpm.h
 *
 *  Copyright (c) 2006-2019 Pacman Development Team <pacman-dev@archlinux.org>
 *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
 *  Copyright (c) 2005 by Aurelien Foret <orelien@chez.com>
 *  Copyright (c) 2005 by Christian Hamar <krics@linuxforum.hu>
 *  Copyright (c) 2005, 2006 by Miklos Vajna <vmiklos@frugalware.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef HBPM_H
#define HBPM_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>    /* int64_t */
#include <sys/types.h> /* off_t */
#include <stdarg.h>    /* va_list */

/* libarchive */
#include <archive.h>
#include <archive_entry.h>

#include <hbpm_list.h>

/*
 * HyperbolaBSD Package Management library
 */

/*
 * Opaque Structures
 */
typedef struct __hbpm_handle_t hbpm_handle_t;
typedef struct __hbpm_db_t hbpm_db_t;
typedef struct __hbpm_pkg_t hbpm_pkg_t;
typedef struct __hbpm_trans_t hbpm_trans_t;

/** @addtogroup hbpm_api_errors Error Codes
 * @{
 */
typedef enum _hbpm_errno_t {
	HBPM_ERR_OK = 0,
	HBPM_ERR_MEMORY,
	HBPM_ERR_SYSTEM,
	HBPM_ERR_BADPERMS,
	HBPM_ERR_NOT_A_FILE,
	HBPM_ERR_NOT_A_DIR,
	HBPM_ERR_WRONG_ARGS,
	HBPM_ERR_DISK_SPACE,
	/* Interface */
	HBPM_ERR_HANDLE_NULL,
	HBPM_ERR_HANDLE_NOT_NULL,
	HBPM_ERR_HANDLE_LOCK,
	/* Databases */
	HBPM_ERR_DB_OPEN,
	HBPM_ERR_DB_CREATE,
	HBPM_ERR_DB_NULL,
	HBPM_ERR_DB_NOT_NULL,
	HBPM_ERR_DB_NOT_FOUND,
	HBPM_ERR_DB_INVALID,
	HBPM_ERR_DB_INVALID_SIG,
	HBPM_ERR_DB_VERSION,
	HBPM_ERR_DB_WRITE,
	HBPM_ERR_DB_REMOVE,
	/* Servers */
	HBPM_ERR_SERVER_BAD_URL,
	HBPM_ERR_SERVER_NONE,
	/* Transactions */
	HBPM_ERR_TRANS_NOT_NULL,
	HBPM_ERR_TRANS_NULL,
	HBPM_ERR_TRANS_DUP_TARGET,
	HBPM_ERR_TRANS_NOT_INITIALIZED,
	HBPM_ERR_TRANS_NOT_PREPARED,
	HBPM_ERR_TRANS_ABORT,
	HBPM_ERR_TRANS_TYPE,
	HBPM_ERR_TRANS_NOT_LOCKED,
	HBPM_ERR_TRANS_HOOK_FAILED,
	/* Packages */
	HBPM_ERR_PKG_NOT_FOUND,
	HBPM_ERR_PKG_IGNORED,
	HBPM_ERR_PKG_INVALID,
	HBPM_ERR_PKG_INVALID_CHECKSUM,
	HBPM_ERR_PKG_INVALID_SIG,
	HBPM_ERR_PKG_MISSING_SIG,
	HBPM_ERR_PKG_OPEN,
	HBPM_ERR_PKG_CANT_REMOVE,
	HBPM_ERR_PKG_INVALID_NAME,
	HBPM_ERR_PKG_INVALID_ARCH,
	HBPM_ERR_PKG_REPO_NOT_FOUND,
	/* Signatures */
	HBPM_ERR_SIG_MISSING,
	HBPM_ERR_SIG_INVALID,
	/* Dependencies */
	HBPM_ERR_UNSATISFIED_DEPS,
	HBPM_ERR_CONFLICTING_DEPS,
	HBPM_ERR_FILE_CONFLICTS,
	/* Misc */
	HBPM_ERR_RETRIEVE,
	HBPM_ERR_INVALID_REGEX,
	/* External library errors */
	HBPM_ERR_LIBARCHIVE,
	HBPM_ERR_LIBCURL,
	HBPM_ERR_EXTERNAL_DOWNLOAD,
	HBPM_ERR_GPGME,
	/* Missing compile-time features */
	HBPM_ERR_MISSING_CAPABILITY_SIGNATURES
} hbpm_errno_t;

/** Returns the current error code from the handle. */
hbpm_errno_t hbpm_errno(hbpm_handle_t *handle);

/** Returns the string corresponding to an error number. */
const char *hbpm_strerror(hbpm_errno_t err);

/* End of hbpm_api_errors */
/** @} */

/** @addtogroup hbpm_api Public API
 * The libhbpm Public API
 * @{
 */

typedef int64_t hbpm_time_t;

/*
 * Enumerations
 * These ones are used in multiple contexts, so are forward-declared.
 */

/** Package install reasons. */
typedef enum _hbpm_pkgreason_t {
	/** Explicitly requested by the user. */
	HBPM_PKG_REASON_EXPLICIT = 0,
	/** Installed as a dependency for another package. */
	HBPM_PKG_REASON_DEPEND = 1
} hbpm_pkgreason_t;

/** Location a package object was loaded from. */
typedef enum _hbpm_pkgfrom_t {
	HBPM_PKG_FROM_FILE = 1,
	HBPM_PKG_FROM_LOCALDB,
	HBPM_PKG_FROM_SYNCDB
} hbpm_pkgfrom_t;

/** Method used to validate a package. */
typedef enum _hbpm_pkgvalidation_t {
	HBPM_PKG_VALIDATION_UNKNOWN = 0,
	HBPM_PKG_VALIDATION_NONE = (1 << 0),
	HBPM_PKG_VALIDATION_MD5SUM = (1 << 1),
	HBPM_PKG_VALIDATION_SHA256SUM = (1 << 2),
	HBPM_PKG_VALIDATION_SIGNATURE = (1 << 3)
} hbpm_pkgvalidation_t;

/** Types of version constraints in dependency specs. */
typedef enum _hbpm_depmod_t {
	/** No version constraint */
	HBPM_DEP_MOD_ANY = 1,
	/** Test version equality (package=x.y.z) */
	HBPM_DEP_MOD_EQ,
	/** Test for at least a version (package>=x.y.z) */
	HBPM_DEP_MOD_GE,
	/** Test for at most a version (package<=x.y.z) */
	HBPM_DEP_MOD_LE,
	/** Test for greater than some version (package>x.y.z) */
	HBPM_DEP_MOD_GT,
	/** Test for less than some version (package<x.y.z) */
	HBPM_DEP_MOD_LT
} hbpm_depmod_t;

/**
 * File conflict type.
 * Whether the conflict results from a file existing on the filesystem, or with
 * another target in the transaction.
 */
typedef enum _hbpm_fileconflicttype_t {
	HBPM_FILECONFLICT_TARGET = 1,
	HBPM_FILECONFLICT_FILESYSTEM
} hbpm_fileconflicttype_t;

/** PGP signature verification options */
typedef enum _hbpm_siglevel_t {
	HBPM_SIG_PACKAGE = (1 << 0),
	HBPM_SIG_PACKAGE_OPTIONAL = (1 << 1),
	HBPM_SIG_PACKAGE_MARGINAL_OK = (1 << 2),
	HBPM_SIG_PACKAGE_UNKNOWN_OK = (1 << 3),

	HBPM_SIG_DATABASE = (1 << 10),
	HBPM_SIG_DATABASE_OPTIONAL = (1 << 11),
	HBPM_SIG_DATABASE_MARGINAL_OK = (1 << 12),
	HBPM_SIG_DATABASE_UNKNOWN_OK = (1 << 13),

	HBPM_SIG_USE_DEFAULT = (1 << 30)
} hbpm_siglevel_t;

/** PGP signature verification status return codes */
typedef enum _hbpm_sigstatus_t {
	HBPM_SIGSTATUS_VALID,
	HBPM_SIGSTATUS_KEY_EXPIRED,
	HBPM_SIGSTATUS_SIG_EXPIRED,
	HBPM_SIGSTATUS_KEY_UNKNOWN,
	HBPM_SIGSTATUS_KEY_DISABLED,
	HBPM_SIGSTATUS_INVALID
} hbpm_sigstatus_t;

/** PGP signature verification status return codes */
typedef enum _hbpm_sigvalidity_t {
	HBPM_SIGVALIDITY_FULL,
	HBPM_SIGVALIDITY_MARGINAL,
	HBPM_SIGVALIDITY_NEVER,
	HBPM_SIGVALIDITY_UNKNOWN
} hbpm_sigvalidity_t;

/*
 * Structures
 */

/** Dependency */
typedef struct _hbpm_depend_t {
	char *name;
	char *version;
	char *desc;
	unsigned long name_hash;
	hbpm_depmod_t mod;
} hbpm_depend_t;

/** Missing dependency */
typedef struct _hbpm_depmissing_t {
	char *target;
	hbpm_depend_t *depend;
	/* this is used only in the case of a remove dependency error */
	char *causingpkg;
} hbpm_depmissing_t;

/** Conflict */
typedef struct _hbpm_conflict_t {
	unsigned long package1_hash;
	unsigned long package2_hash;
	char *package1;
	char *package2;
	hbpm_depend_t *reason;
} hbpm_conflict_t;

/** File conflict */
typedef struct _hbpm_fileconflict_t {
	char *target;
	hbpm_fileconflicttype_t type;
	char *file;
	char *ctarget;
} hbpm_fileconflict_t;

/** Package group */
typedef struct _hbpm_group_t {
	/** group name */
	char *name;
	/** list of hbpm_pkg_t packages */
	hbpm_list_t *packages;
} hbpm_group_t;

/** File in a package */
typedef struct _hbpm_file_t {
	char *name;
	off_t size;
	mode_t mode;
} hbpm_file_t;

/** Package filelist container */
typedef struct _hbpm_filelist_t {
	size_t count;
	hbpm_file_t *files;
} hbpm_filelist_t;

/** Local package or package file backup entry */
typedef struct _hbpm_backup_t {
	char *name;
	char *hash;
} hbpm_backup_t;

typedef struct _hbpm_pgpkey_t {
	void *data;
	char *fingerprint;
	char *uid;
	char *name;
	char *email;
	hbpm_time_t created;
	hbpm_time_t expires;
	unsigned int length;
	unsigned int revoked;
	char pubkey_algo;
} hbpm_pgpkey_t;

/**
 * Signature result. Contains the key, status, and validity of a given
 * signature.
 */
typedef struct _hbpm_sigresult_t {
	hbpm_pgpkey_t key;
	hbpm_sigstatus_t status;
	hbpm_sigvalidity_t validity;
} hbpm_sigresult_t;

/**
 * Signature list. Contains the number of signatures found and a pointer to an
 * array of results. The array is of size count.
 */
typedef struct _hbpm_siglist_t {
	size_t count;
	hbpm_sigresult_t *results;
} hbpm_siglist_t;


/*
 * Hooks
 */

typedef enum _hbpm_hook_when_t {
	HBPM_HOOK_PRE_TRANSACTION = 1,
	HBPM_HOOK_POST_TRANSACTION
} hbpm_hook_when_t;

/*
 * Logging facilities
 */

/** \addtogroup hbmp_log Logging Functions
 * @brief Functions to log using libhbpm
 * @{
 */

/** Logging Levels */
typedef enum _hbpm_loglevel_t {
	HBPM_LOG_ERROR    = 1,
	HBPM_LOG_WARNING  = (1 << 1),
	HBPM_LOG_DEBUG    = (1 << 2),
	HBPM_LOG_FUNCTION = (1 << 3)
} hbpm_loglevel_t;

typedef void (*hbpm_cb_log)(hbpm_loglevel_t, const char *, va_list);

/** A printf-like function for logging.
 * @param handle the context handle
 * @param prefix caller-specific prefix for the log
 * @param fmt output format
 * @return 0 on success, -1 on error (pm_errno is set accordingly)
 */
int hbpm_logaction(hbpm_handle_t *handle, const char *prefix,
		const char *fmt, ...) __attribute__((format(printf, 3, 4)));

/** @} */

/**
 * Type of events.
 */
typedef enum _hbpm_event_type_t {
	/** Dependencies will be computed for a package. */
	HBPM_EVENT_CHECKDEPS_START = 1,
	/** Dependencies were computed for a package. */
	HBPM_EVENT_CHECKDEPS_DONE,
	/** File conflicts will be computed for a package. */
	HBPM_EVENT_FILECONFLICTS_START,
	/** File conflicts were computed for a package. */
	HBPM_EVENT_FILECONFLICTS_DONE,
	/** Dependencies will be resolved for target package. */
	HBPM_EVENT_RESOLVEDEPS_START,
	/** Dependencies were resolved for target package. */
	HBPM_EVENT_RESOLVEDEPS_DONE,
	/** Inter-conflicts will be checked for target package. */
	HBPM_EVENT_INTERCONFLICTS_START,
	/** Inter-conflicts were checked for target package. */
	HBPM_EVENT_INTERCONFLICTS_DONE,
	/** Processing the package transaction is starting. */
	HBPM_EVENT_TRANSACTION_START,
	/** Processing the package transaction is finished. */
	HBPM_EVENT_TRANSACTION_DONE,
	/** Package will be installed/upgraded/downgraded/re-installed/removed; See
	 * hbpm_event_package_operation_t for arguments. */
	HBPM_EVENT_PACKAGE_OPERATION_START,
	/** Package was installed/upgraded/downgraded/re-installed/removed; See
	 * hbpm_event_package_operation_t for arguments. */
	HBPM_EVENT_PACKAGE_OPERATION_DONE,
	/** Target package's integrity will be checked. */
	HBPM_EVENT_INTEGRITY_START,
	/** Target package's integrity was checked. */
	HBPM_EVENT_INTEGRITY_DONE,
	/** Target package will be loaded. */
	HBPM_EVENT_LOAD_START,
	/** Target package is finished loading. */
	HBPM_EVENT_LOAD_DONE,
	/** Scriptlet has printed information; See hbpm_event_scriptlet_info_t for
	 * arguments. */
	HBPM_EVENT_SCRIPTLET_INFO,
	/** Files will be downloaded from a repository. */
	HBPM_EVENT_RETRIEVE_START,
	/** Files were downloaded from a repository. */
	HBPM_EVENT_RETRIEVE_DONE,
	/** Not all files were successfully downloaded from a repository. */
	HBPM_EVENT_RETRIEVE_FAILED,
	/** A file will be downloaded from a repository; See hbpm_event_pkgdownload_t
	 * for arguments */
	HBPM_EVENT_PKGDOWNLOAD_START,
	/** A file was downloaded from a repository; See hbpm_event_pkgdownload_t
	 * for arguments */
	HBPM_EVENT_PKGDOWNLOAD_DONE,
	/** A file failed to be downloaded from a repository; See
	 * hbpm_event_pkgdownload_t for arguments */
	HBPM_EVENT_PKGDOWNLOAD_FAILED,
	/** Disk space usage will be computed for a package. */
	HBPM_EVENT_DISKSPACE_START,
	/** Disk space usage was computed for a package. */
	HBPM_EVENT_DISKSPACE_DONE,
	/** An optdepend for another package is being removed; See
	 * hbpm_event_optdep_removal_t for arguments. */
	HBPM_EVENT_OPTDEP_REMOVAL,
	/** A configured repository database is missing; See
	 * hbpm_event_database_missing_t for arguments. */
	HBPM_EVENT_DATABASE_MISSING,
	/** Checking keys used to create signatures are in keyring. */
	HBPM_EVENT_KEYRING_START,
	/** Keyring checking is finished. */
	HBPM_EVENT_KEYRING_DONE,
	/** Downloading missing keys into keyring. */
	HBPM_EVENT_KEY_DOWNLOAD_START,
	/** Key downloading is finished. */
	HBPM_EVENT_KEY_DOWNLOAD_DONE,
	/** A .hypernew file was created; See hbpm_event_hypernew_created_t for arguments. */
	HBPM_EVENT_HYPERNEW_CREATED,
	/** A .hypersave file was created; See hbpm_event_hypersave_created_t for
	 * arguments */
	HBPM_EVENT_HYPERSAVE_CREATED,
	/** Processing hooks will be started. */
	HBPM_EVENT_HOOK_START,
	/** Processing hooks is finished. */
	HBPM_EVENT_HOOK_DONE,
	/** A hook is starting */
	HBPM_EVENT_HOOK_RUN_START,
	/** A hook has finished running */
	HBPM_EVENT_HOOK_RUN_DONE
} hbpm_event_type_t;

typedef struct _hbpm_event_any_t {
	/** Type of event. */
	hbpm_event_type_t type;
} hbpm_event_any_t;

typedef enum _hbpm_package_operation_t {
	/** Package (to be) installed. (No oldpkg) */
	HBPM_PACKAGE_INSTALL = 1,
	/** Package (to be) upgraded */
	HBPM_PACKAGE_UPGRADE,
	/** Package (to be) re-installed. */
	HBPM_PACKAGE_REINSTALL,
	/** Package (to be) downgraded. */
	HBPM_PACKAGE_DOWNGRADE,
	/** Package (to be) removed. (No newpkg) */
	HBPM_PACKAGE_REMOVE
} hbpm_package_operation_t;

typedef struct _hbpm_event_package_operation_t {
	/** Type of event. */
	hbpm_event_type_t type;
	/** Type of operation. */
	hbpm_package_operation_t operation;
	/** Old package. */
	hbpm_pkg_t *oldpkg;
	/** New package. */
	hbpm_pkg_t *newpkg;
} hbpm_event_package_operation_t;

typedef struct _hbpm_event_optdep_removal_t {
	/** Type of event. */
	hbpm_event_type_t type;
	/** Package with the optdep. */
	hbpm_pkg_t *pkg;
	/** Optdep being removed. */
	hbpm_depend_t *optdep;
} hbpm_event_optdep_removal_t;

typedef struct _hbpm_event_scriptlet_info_t {
	/** Type of event. */
	hbpm_event_type_t type;
	/** Line of scriptlet output. */
	const char *line;
} hbpm_event_scriptlet_info_t;

typedef struct _hbpm_event_database_missing_t {
	/** Type of event. */
	hbpm_event_type_t type;
	/** Name of the database. */
	const char *dbname;
} hbpm_event_database_missing_t;

typedef struct _hbpm_event_pkgdownload_t {
	/** Type of event. */
	hbpm_event_type_t type;
	/** Name of the file */
	const char *file;
} hbpm_event_pkgdownload_t;

typedef struct _hbpm_event_hypernew_created_t {
	/** Type of event. */
	hbpm_event_type_t type;
	/** Whether the creation was result of a NoUpgrade or not */
	int from_noupgrade;
	/** Old package. */
	hbpm_pkg_t *oldpkg;
	/** New Package. */
	hbpm_pkg_t *newpkg;
	/** Filename of the file without the .hypernew suffix */
	const char *file;
} hbpm_event_hypernew_created_t;

typedef struct _hbpm_event_hypersave_created_t {
	/** Type of event. */
	hbpm_event_type_t type;
	/** Old package. */
	hbpm_pkg_t *oldpkg;
	/** Filename of the file without the .hypersave suffix. */
	const char *file;
} hbpm_event_hypersave_created_t;

typedef struct _hbpm_event_hook_t {
	/** Type of event.*/
	hbpm_event_type_t type;
	/** Type of hooks. */
	hbpm_hook_when_t when;
} hbpm_event_hook_t;

typedef struct _hbpm_event_hook_run_t {
	/** Type of event.*/
	hbpm_event_type_t type;
	/** Name of hook */
	const char *name;
	/** Description of hook to be outputted */
	const char *desc;
	/** position of hook being run */
	size_t position;
	/** total hooks being run */
	size_t total;
} hbpm_event_hook_run_t;

/** Events.
 * This is an union passed to the callback, that allows the frontend to know
 * which type of event was triggered (via type). It is then possible to
 * typecast the pointer to the right structure, or use the union field, in order
 * to access event-specific data. */
typedef union _hbpm_event_t {
	hbpm_event_type_t type;
	hbpm_event_any_t any;
	hbpm_event_package_operation_t package_operation;
	hbpm_event_optdep_removal_t optdep_removal;
	hbpm_event_scriptlet_info_t scriptlet_info;
	hbpm_event_database_missing_t database_missing;
	hbpm_event_pkgdownload_t pkgdownload;
	hbpm_event_hypernew_created_t hypernew_created;
	hbpm_event_hypersave_created_t hypersave_created;
	hbpm_event_hook_t hook;
	hbpm_event_hook_run_t hook_run;
} hbpm_event_t;

/** Event callback. */
typedef void (*hbpm_cb_event)(hbpm_event_t *);

/**
 * Type of questions.
 * Unlike the events or progress enumerations, this enum has bitmask values
 * so a frontend can use a bitmask map to supply preselected answers to the
 * different types of questions.
 */
typedef enum _hbpm_question_type_t {
	HBPM_QUESTION_INSTALL_IGNOREPKG = (1 << 0),
	HBPM_QUESTION_REPLACE_PKG = (1 << 1),
	HBPM_QUESTION_CONFLICT_PKG = (1 << 2),
	HBPM_QUESTION_CORRUPTED_PKG = (1 << 3),
	HBPM_QUESTION_REMOVE_PKGS = (1 << 4),
	HBPM_QUESTION_SELECT_PROVIDER = (1 << 5),
	HBPM_QUESTION_IMPORT_KEY = (1 << 6)
} hbpm_question_type_t;

typedef struct _hbpm_question_any_t {
	/** Type of question. */
	hbpm_question_type_t type;
	/** Answer. */
	int answer;
} hbpm_question_any_t;

typedef struct _hbpm_question_install_ignorepkg_t {
	/** Type of question. */
	hbpm_question_type_t type;
	/** Answer: whether or not to install pkg anyway. */
	int install;
	/* Package in IgnorePkg/IgnoreGroup. */
	hbpm_pkg_t *pkg;
} hbpm_question_install_ignorepkg_t;

typedef struct _hbpm_question_replace_t {
	/** Type of question. */
	hbpm_question_type_t type;
	/** Answer: whether or not to replace oldpkg with newpkg. */
	int replace;
	/* Package to be replaced. */
	hbpm_pkg_t *oldpkg;
	/* Package to replace with. */
	hbpm_pkg_t *newpkg;
	/* DB of newpkg */
	hbpm_db_t *newdb;
} hbpm_question_replace_t;

typedef struct _hbpm_question_conflict_t {
	/** Type of question. */
	hbpm_question_type_t type;
	/** Answer: whether or not to remove conflict->package2. */
	int remove;
	/** Conflict info. */
	hbpm_conflict_t *conflict;
} hbpm_question_conflict_t;

typedef struct _hbpm_question_corrupted_t {
	/** Type of question. */
	hbpm_question_type_t type;
	/** Answer: whether or not to remove filepath. */
	int remove;
	/** Filename to remove */
	const char *filepath;
	/** Error code indicating the reason for package invalidity */
	hbpm_errno_t reason;
} hbpm_question_corrupted_t;

typedef struct _hbpm_question_remove_pkgs_t {
	/** Type of question. */
	hbpm_question_type_t type;
	/** Answer: whether or not to skip packages. */
	int skip;
	/** List of hbpm_pkg_t* with unresolved dependencies. */
	hbpm_list_t *packages;
} hbpm_question_remove_pkgs_t;

typedef struct _hbpm_question_select_provider_t {
	/** Type of question. */
	hbpm_question_type_t type;
	/** Answer: which provider to use (index from providers). */
	int use_index;
	/** List of hbpm_pkg_t* as possible providers. */
	hbpm_list_t *providers;
	/** What providers provide for. */
	hbpm_depend_t *depend;
} hbpm_question_select_provider_t;

typedef struct _hbpm_question_import_key_t {
	/** Type of question. */
	hbpm_question_type_t type;
	/** Answer: whether or not to import key. */
	int import;
	/** The key to import. */
	hbpm_pgpkey_t *key;
} hbpm_question_import_key_t;

/**
 * Questions.
 * This is an union passed to the callback, that allows the frontend to know
 * which type of question was triggered (via type). It is then possible to
 * typecast the pointer to the right structure, or use the union field, in order
 * to access question-specific data. */
typedef union _hbpm_question_t {
	hbpm_question_type_t type;
	hbpm_question_any_t any;
	hbpm_question_install_ignorepkg_t install_ignorepkg;
	hbpm_question_replace_t replace;
	hbpm_question_conflict_t conflict;
	hbpm_question_corrupted_t corrupted;
	hbpm_question_remove_pkgs_t remove_pkgs;
	hbpm_question_select_provider_t select_provider;
	hbpm_question_import_key_t import_key;
} hbpm_question_t;

/** Question callback */
typedef void (*hbpm_cb_question)(hbpm_question_t *);

/** Progress */
typedef enum _hbpm_progress_t {
	HBPM_PROGRESS_ADD_START,
	HBPM_PROGRESS_UPGRADE_START,
	HBPM_PROGRESS_DOWNGRADE_START,
	HBPM_PROGRESS_REINSTALL_START,
	HBPM_PROGRESS_REMOVE_START,
	HBPM_PROGRESS_CONFLICTS_START,
	HBPM_PROGRESS_DISKSPACE_START,
	HBPM_PROGRESS_INTEGRITY_START,
	HBPM_PROGRESS_LOAD_START,
	HBPM_PROGRESS_KEYRING_START
} hbpm_progress_t;

/** Progress callback */
typedef void (*hbpm_cb_progress)(hbpm_progress_t, const char *, int, size_t, size_t);

/*
 * Downloading
 */

/** Type of download progress callbacks.
 * @param filename the name of the file being downloaded
 * @param xfered the number of transferred bytes
 * @param total the total number of bytes to transfer
 */
typedef void (*hbpm_cb_download)(const char *filename,
		off_t xfered, off_t total);

typedef void (*hbpm_cb_totaldl)(off_t total);

/** A callback for downloading files
 * @param url the URL of the file to be downloaded
 * @param localpath the directory to which the file should be downloaded
 * @param force whether to force an update, even if the file is the same
 * @return 0 on success, 1 if the file exists and is identical, -1 on
 * error.
 */
typedef int (*hbpm_cb_fetch)(const char *url, const char *localpath,
		int force);

/** Fetch a remote pkg.
 * @param handle the context handle
 * @param url URL of the package to download
 * @return the downloaded filepath on success, NULL on error
 */
char *hbpm_fetch_pkgurl(hbpm_handle_t *handle, const char *url);

/** @addtogroup hbpm_api_options Options
 * Libhbpm option getters and setters
 * @{
 */

/** Returns the callback used for logging. */
hbpm_cb_log hbpm_option_get_logcb(hbpm_handle_t *handle);
/** Sets the callback used for logging. */
int hbpm_option_set_logcb(hbpm_handle_t *handle, hbpm_cb_log cb);

/** Returns the callback used to report download progress. */
hbpm_cb_download hbpm_option_get_dlcb(hbpm_handle_t *handle);
/** Sets the callback used to report download progress. */
int hbpm_option_set_dlcb(hbpm_handle_t *handle, hbpm_cb_download cb);

/** Returns the downloading callback. */
hbpm_cb_fetch hbpm_option_get_fetchcb(hbpm_handle_t *handle);
/** Sets the downloading callback. */
int hbpm_option_set_fetchcb(hbpm_handle_t *handle, hbpm_cb_fetch cb);

/** Returns the callback used to report total download size. */
hbpm_cb_totaldl hbpm_option_get_totaldlcb(hbpm_handle_t *handle);
/** Sets the callback used to report total download size. */
int hbpm_option_set_totaldlcb(hbpm_handle_t *handle, hbpm_cb_totaldl cb);

/** Returns the callback used for events. */
hbpm_cb_event hbpm_option_get_eventcb(hbpm_handle_t *handle);
/** Sets the callback used for events. */
int hbpm_option_set_eventcb(hbpm_handle_t *handle, hbpm_cb_event cb);

/** Returns the callback used for questions. */
hbpm_cb_question hbpm_option_get_questioncb(hbpm_handle_t *handle);
/** Sets the callback used for questions. */
int hbpm_option_set_questioncb(hbpm_handle_t *handle, hbpm_cb_question cb);

/** Returns the callback used for operation progress. */
hbpm_cb_progress hbpm_option_get_progresscb(hbpm_handle_t *handle);
/** Sets the callback used for operation progress. */
int hbpm_option_set_progresscb(hbpm_handle_t *handle, hbpm_cb_progress cb);

/** Returns the root of the destination filesystem. Read-only. */
const char *hbpm_option_get_root(hbpm_handle_t *handle);

/** Returns the path to the database directory. Read-only. */
const char *hbpm_option_get_dbpath(hbpm_handle_t *handle);

/** Get the name of the database lock file. Read-only. */
const char *hbpm_option_get_lockfile(hbpm_handle_t *handle);

/** @name Accessors to the list of package cache directories.
 * @{
 */
hbpm_list_t *hbpm_option_get_cachedirs(hbpm_handle_t *handle);
int hbpm_option_set_cachedirs(hbpm_handle_t *handle, hbpm_list_t *cachedirs);
int hbpm_option_add_cachedir(hbpm_handle_t *handle, const char *cachedir);
int hbpm_option_remove_cachedir(hbpm_handle_t *handle, const char *cachedir);
/** @} */

/** @name Accessors to the list of package hook directories.
 * @{
 */
hbpm_list_t *hbpm_option_get_hookdirs(hbpm_handle_t *handle);
int hbpm_option_set_hookdirs(hbpm_handle_t *handle, hbpm_list_t *hookdirs);
int hbpm_option_add_hookdir(hbpm_handle_t *handle, const char *hookdir);
int hbpm_option_remove_hookdir(hbpm_handle_t *handle, const char *hookdir);
/** @} */

hbpm_list_t *hbpm_option_get_overwrite_files(hbpm_handle_t *handle);
int hbpm_option_set_overwrite_files(hbpm_handle_t *handle, hbpm_list_t *globs);
int hbpm_option_add_overwrite_file(hbpm_handle_t *handle, const char *glob);
int hbpm_option_remove_overwrite_file(hbpm_handle_t *handle, const char *glob);

/** Returns the logfile name. */
const char *hbpm_option_get_logfile(hbpm_handle_t *handle);
/** Sets the logfile name. */
int hbpm_option_set_logfile(hbpm_handle_t *handle, const char *logfile);

/** Returns the path to libhbpm's GnuPG home directory. */
const char *hbpm_option_get_gpgdir(hbpm_handle_t *handle);
/** Sets the path to libhbpm's GnuPG home directory. */
int hbpm_option_set_gpgdir(hbpm_handle_t *handle, const char *gpgdir);

/** Returns whether to use syslog (0 is FALSE, TRUE otherwise). */
int hbpm_option_get_usesyslog(hbpm_handle_t *handle);
/** Sets whether to use syslog (0 is FALSE, TRUE otherwise). */
int hbpm_option_set_usesyslog(hbpm_handle_t *handle, int usesyslog);

/** @name Accessors to the list of no-upgrade files.
 * These functions modify the list of files which should
 * not be updated by package installation.
 * @{
 */
hbpm_list_t *hbpm_option_get_noupgrades(hbpm_handle_t *handle);
int hbpm_option_add_noupgrade(hbpm_handle_t *handle, const char *path);
int hbpm_option_set_noupgrades(hbpm_handle_t *handle, hbpm_list_t *noupgrade);
int hbpm_option_remove_noupgrade(hbpm_handle_t *handle, const char *path);
int hbpm_option_match_noupgrade(hbpm_handle_t *handle, const char *path);
/** @} */

/** @name Accessors to the list of no-extract files.
 * These functions modify the list of filenames which should
 * be skipped packages which should
 * not be upgraded by a sysupgrade operation.
 * @{
 */
hbpm_list_t *hbpm_option_get_noextracts(hbpm_handle_t *handle);
int hbpm_option_add_noextract(hbpm_handle_t *handle, const char *path);
int hbpm_option_set_noextracts(hbpm_handle_t *handle, hbpm_list_t *noextract);
int hbpm_option_remove_noextract(hbpm_handle_t *handle, const char *path);
int hbpm_option_match_noextract(hbpm_handle_t *handle, const char *path);
/** @} */

/** @name Accessors to the list of ignored packages.
 * These functions modify the list of packages that
 * should be ignored by a sysupgrade.
 * @{
 */
hbpm_list_t *hbpm_option_get_ignorepkgs(hbpm_handle_t *handle);
int hbpm_option_add_ignorepkg(hbpm_handle_t *handle, const char *pkg);
int hbpm_option_set_ignorepkgs(hbpm_handle_t *handle, hbpm_list_t *ignorepkgs);
int hbpm_option_remove_ignorepkg(hbpm_handle_t *handle, const char *pkg);
/** @} */

/** @name Accessors to the list of ignored groups.
 * These functions modify the list of groups whose packages
 * should be ignored by a sysupgrade.
 * @{
 */
hbpm_list_t *hbpm_option_get_ignoregroups(hbpm_handle_t *handle);
int hbpm_option_add_ignoregroup(hbpm_handle_t *handle, const char *grp);
int hbpm_option_set_ignoregroups(hbpm_handle_t *handle, hbpm_list_t *ignoregrps);
int hbpm_option_remove_ignoregroup(hbpm_handle_t *handle, const char *grp);
/** @} */

/** @name Accessors to the list of ignored dependencies.
 * These functions modify the list of dependencies that
 * should be ignored by a sysupgrade.
 * @{
 */
hbpm_list_t *hbpm_option_get_assumeinstalled(hbpm_handle_t *handle);
int hbpm_option_add_assumeinstalled(hbpm_handle_t *handle, const hbpm_depend_t *dep);
int hbpm_option_set_assumeinstalled(hbpm_handle_t *handle, hbpm_list_t *deps);
int hbpm_option_remove_assumeinstalled(hbpm_handle_t *handle, const hbpm_depend_t *dep);
/** @} */

/** Returns the targeted architecture. */
const char *hbpm_option_get_arch(hbpm_handle_t *handle);
/** Sets the targeted architecture. */
int hbpm_option_set_arch(hbpm_handle_t *handle, const char *arch);

int hbpm_option_get_checkspace(hbpm_handle_t *handle);
int hbpm_option_set_checkspace(hbpm_handle_t *handle, int checkspace);

const char *hbpm_option_get_dbext(hbpm_handle_t *handle);
int hbpm_option_set_dbext(hbpm_handle_t *handle, const char *dbext);

int hbpm_option_get_default_siglevel(hbpm_handle_t *handle);
int hbpm_option_set_default_siglevel(hbpm_handle_t *handle, int level);

int hbpm_option_get_local_file_siglevel(hbpm_handle_t *handle);
int hbpm_option_set_local_file_siglevel(hbpm_handle_t *handle, int level);

int hbpm_option_get_remote_file_siglevel(hbpm_handle_t *handle);
int hbpm_option_set_remote_file_siglevel(hbpm_handle_t *handle, int level);

int hbpm_option_set_disable_dl_timeout(hbpm_handle_t *handle, unsigned short disable_dl_timeout);

/** @} */

/** @addtogroup hbpm_api_databases Database Functions
 * Functions to query and manipulate the database of libhbpm.
 * @{
 */

/** Get the database of locally installed packages.
 * The returned pointer points to an internal structure
 * of libhbpm which should only be manipulated through
 * libhbpm functions.
 * @return a reference to the local database
 */
hbpm_db_t *hbpm_get_localdb(hbpm_handle_t *handle);

/** Get the list of sync databases.
 * Returns a list of hbpm_db_t structures, one for each registered
 * sync database.
 * @param handle the context handle
 * @return a reference to an internal list of hbpm_db_t structures
 */
hbpm_list_t *hbpm_get_syncdbs(hbpm_handle_t *handle);

/** Register a sync database of packages.
 * @param handle the context handle
 * @param treename the name of the sync repository
 * @param level what level of signature checking to perform on the
 * database; note that this must be a '.sig' file type verification
 * @return an hbpm_db_t* on success (the value), NULL on error
 */
hbpm_db_t *hbpm_register_syncdb(hbpm_handle_t *handle, const char *treename,
		int level);

/** Unregister all package databases.
 * @param handle the context handle
 * @return 0 on success, -1 on error (pm_errno is set accordingly)
 */
int hbpm_unregister_all_syncdbs(hbpm_handle_t *handle);

/** Unregister a package database.
 * @param db pointer to the package database to unregister
 * @return 0 on success, -1 on error (pm_errno is set accordingly)
 */
int hbpm_db_unregister(hbpm_db_t *db);

/** Get the name of a package database.
 * @param db pointer to the package database
 * @return the name of the package database, NULL on error
 */
const char *hbpm_db_get_name(const hbpm_db_t *db);

/** Get the signature verification level for a database.
 * Will return the default verification level if this database is set up
 * with HBPM_SIG_USE_DEFAULT.
 * @param db pointer to the package database
 * @return the signature verification level
 */
int hbpm_db_get_siglevel(hbpm_db_t *db);

/** Check the validity of a database.
 * This is most useful for sync databases and verifying signature status.
 * If invalid, the handle error code will be set accordingly.
 * @param db pointer to the package database
 * @return 0 if valid, -1 if invalid (pm_errno is set accordingly)
 */
int hbpm_db_get_valid(hbpm_db_t *db);

/** @name Accessors to the list of servers for a database.
 * @{
 */

/** Get the list of servers assigned to this db.
 * @param db pointer to the database to get the servers from
 * @return a char* list of servers
 */
hbpm_list_t *hbpm_db_get_servers(const hbpm_db_t *db);

/** Sets the list of servers for the database to use.
 * @param db the database to set the servers
 * @param a char* list of servers. Note: the database will
 * take ownership of the list and it should no longer be
 * freed by the caller
 */
int hbpm_db_set_servers(hbpm_db_t *db, hbpm_list_t *servers);

/** Add a download server to a database.
 * @param db database pointer
 * @param url url of the server
 * @return 0 on success, -1 on error (pm_errno is set accordingly)
 */
int hbpm_db_add_server(hbpm_db_t *db, const char *url);

/** Remove a download server from a database.
 * @param db database pointer
 * @param url url of the server
 * @return 0 on success, 1 on server not present,
 * -1 on error (pm_errno is set accordingly)
 */
int hbpm_db_remove_server(hbpm_db_t *db, const char *url);
/** @} */

/** Update a package database
 *
 * An update of the package database \a db will be attempted. Unless
 * \a force is true, the update will only be performed if the remote
 * database was modified since the last update.
 *
 * This operation requires a database lock, and will return an applicable error
 * if the lock could not be obtained.
 *
 * Example:
 * @code
 * hbpm_list_t *syncs = hbpm_get_syncdbs();
 * for(i = syncs; i; i = hbpm_list_next(i)) {
 *     hbpm_db_t *db = hbpm_list_getdata(i);
 *     result = hbpm_db_update(0, db);
 *
 *     if(result < 0) {
 *        printf("Unable to update database: %s\n", hbpm_strerrorlast());
 *     } else if(result == 1) {
 *         printf("Database already up to date\n");
 *     } else {
 *         printf("Database updated\n");
 *     }
 * }
 * @endcode
 *
 * @note After a successful update, the \link hbpm_db_get_pkgcache()
 * package cache \endlink will be invalidated
 * @param force if true, then forces the update, otherwise update only in case
 * the database isn't up to date
 * @param db pointer to the package database to update
 * @return 0 on success, -1 on error (pm_errno is set accordingly), 1 if up to
 * to date
 */
int hbpm_db_update(int force, hbpm_db_t *db);

/** Get a package entry from a package database.
 * @param db pointer to the package database to get the package from
 * @param name of the package
 * @return the package entry on success, NULL on error
 */
hbpm_pkg_t *hbpm_db_get_pkg(hbpm_db_t *db, const char *name);

/** Get the package cache of a package database.
 * @param db pointer to the package database to get the package from
 * @return the list of packages on success, NULL on error
 */
hbpm_list_t *hbpm_db_get_pkgcache(hbpm_db_t *db);

/** Get a group entry from a package database.
 * @param db pointer to the package database to get the group from
 * @param name of the group
 * @return the groups entry on success, NULL on error
 */
hbpm_group_t *hbpm_db_get_group(hbpm_db_t *db, const char *name);

/** Get the group cache of a package database.
 * @param db pointer to the package database to get the group from
 * @return the list of groups on success, NULL on error
 */
hbpm_list_t *hbpm_db_get_groupcache(hbpm_db_t *db);

/** Searches a database with regular expressions.
 * @param db pointer to the package database to search in
 * @param needles a list of regular expressions to search for
 * @param ret pointer to list for storing packages matching all
 * regular expressions - must point to an empty (NULL) hbpm_list_t *.
 * @return 0 on success, -1 on error (pm_errno is set accordingly)
 */
int hbpm_db_search(hbpm_db_t *db, const hbpm_list_t *needles,
		hbpm_list_t **ret);

typedef enum _hbpm_db_usage_t {
	HBPM_DB_USAGE_SYNC = 1,
	HBPM_DB_USAGE_SEARCH = (1 << 1),
	HBPM_DB_USAGE_INSTALL = (1 << 2),
	HBPM_DB_USAGE_UPGRADE = (1 << 3),
	HBPM_DB_USAGE_ALL = (1 << 4) - 1,
} hbpm_db_usage_t;

/** Sets the usage of a database.
 * @param db pointer to the package database to set the status for
 * @param usage a bitmask of hbpm_db_usage_t values
 * @return 0 on success, or -1 on error
 */
int hbpm_db_set_usage(hbpm_db_t *db, int usage);

/** Gets the usage of a database.
 * @param db pointer to the package database to get the status of
 * @param usage pointer to an hbpm_db_usage_t to store db's status
 * @return 0 on success, or -1 on error
 */
int hbpm_db_get_usage(hbpm_db_t *db, int *usage);

/** @} */

/** @addtogroup hbpm_api_packages Package Functions
 * Functions to manipulate libhbpm packages
 * @{
 */

/** Create a package from a file.
 * If full is false, the archive is read only until all necessary
 * metadata is found. If it is true, the entire archive is read, which
 * serves as a verification of integrity and the filelist can be created.
 * The allocated structure should be freed using hbpm_pkg_free().
 * @param handle the context handle
 * @param filename location of the package tarball
 * @param full whether to stop the load after metadata is read or continue
 * through the full archive
 * @param level what level of package signature checking to perform on the
 * package; note that this must be a '.sig' file type verification
 * @param pkg address of the package pointer
 * @return 0 on success, -1 on error (pm_errno is set accordingly)
 */
int hbpm_pkg_load(hbpm_handle_t *handle, const char *filename, int full,
		int level, hbpm_pkg_t **pkg);

/** Find a package in a list by name.
 * @param haystack a list of hbpm_pkg_t
 * @param needle the package name
 * @return a pointer to the package if found or NULL
 */
hbpm_pkg_t *hbpm_pkg_find(hbpm_list_t *haystack, const char *needle);

/** Free a package.
 * @param pkg package pointer to free
 * @return 0 on success, -1 on error (pm_errno is set accordingly)
 */
int hbpm_pkg_free(hbpm_pkg_t *pkg);

/** Check the integrity (with md5) of a package from the sync cache.
 * @param pkg package pointer
 * @return 0 on success, -1 on error (pm_errno is set accordingly)
 */
int hbpm_pkg_checkmd5sum(hbpm_pkg_t *pkg);

/** Compare two version strings and determine which one is 'newer'.
 * Returns a value comparable to the way strcmp works. Returns 1
 * if a is newer than b, 0 if a and b are the same version, or -1
 * if b is newer than a.
 *
 * Different epoch values for version strings will override any further
 * comparison. If no epoch is provided, 0 is assumed.
 *
 * Keep in mind that the pkgrel is only compared if it is available
 * on both versions handed to this function. For example, comparing
 * 1.5-1 and 1.5 will yield 0; comparing 1.5-1 and 1.5-2 will yield
 * -1 as expected. This is mainly for supporting versioned dependencies
 * that do not include the pkgrel.
 */
int hbpm_pkg_vercmp(const char *a, const char *b);

/** Computes the list of packages requiring a given package.
 * The return value of this function is a newly allocated
 * list of package names (char*), it should be freed by the caller.
 * @param pkg a package
 * @return the list of packages requiring pkg
 */
hbpm_list_t *hbpm_pkg_compute_requiredby(hbpm_pkg_t *pkg);

/** Computes the list of packages optionally requiring a given package.
 * The return value of this function is a newly allocated
 * list of package names (char*), it should be freed by the caller.
 * @param pkg a package
 * @return the list of packages optionally requiring pkg
 */
hbpm_list_t *hbpm_pkg_compute_optionalfor(hbpm_pkg_t *pkg);

/** Test if a package should be ignored.
 * Checks if the package is ignored via IgnorePkg, or if the package is
 * in a group ignored via IgnoreGroup.
 * @param handle the context handle
 * @param pkg the package to test
 * @return 1 if the package should be ignored, 0 otherwise
 */
int hbpm_pkg_should_ignore(hbpm_handle_t *handle, hbpm_pkg_t *pkg);

/** @name Package Property Accessors
 * Any pointer returned by these functions points to internal structures
 * allocated by libhbpm. They should not be freed nor modified in any
 * way.
 * @{
 */

/** Gets the name of the file from which the package was loaded.
 * @param pkg a pointer to package
 * @return a reference to an internal string
 */
const char *hbpm_pkg_get_filename(hbpm_pkg_t *pkg);

/** Returns the package base name.
 * @param pkg a pointer to package
 * @return a reference to an internal string
 */
const char *hbpm_pkg_get_base(hbpm_pkg_t *pkg);

/** Returns the package name.
 * @param pkg a pointer to package
 * @return a reference to an internal string
 */
const char *hbpm_pkg_get_name(hbpm_pkg_t *pkg);

/** Returns the package version as a string.
 * This includes all available epoch, version, and pkgrel components. Use
 * hbpm_pkg_vercmp() to compare version strings if necessary.
 * @param pkg a pointer to package
 * @return a reference to an internal string
 */
const char *hbpm_pkg_get_version(hbpm_pkg_t *pkg);

/** Returns the origin of the package.
 * @return an hbpm_pkgfrom_t constant, -1 on error
 */
hbpm_pkgfrom_t hbpm_pkg_get_origin(hbpm_pkg_t *pkg);

/** Returns the package description.
 * @param pkg a pointer to package
 * @return a reference to an internal string
 */
const char *hbpm_pkg_get_desc(hbpm_pkg_t *pkg);

/** Returns the package URL.
 * @param pkg a pointer to package
 * @return a reference to an internal string
 */
const char *hbpm_pkg_get_url(hbpm_pkg_t *pkg);

/** Returns the build timestamp of the package.
 * @param pkg a pointer to package
 * @return the timestamp of the build time
 */
hbpm_time_t hbpm_pkg_get_builddate(hbpm_pkg_t *pkg);

/** Returns the install timestamp of the package.
 * @param pkg a pointer to package
 * @return the timestamp of the install time
 */
hbpm_time_t hbpm_pkg_get_installdate(hbpm_pkg_t *pkg);

/** Returns the packager's name.
 * @param pkg a pointer to package
 * @return a reference to an internal string
 */
const char *hbpm_pkg_get_packager(hbpm_pkg_t *pkg);

/** Returns the package's MD5 checksum as a string.
 * The returned string is a sequence of 32 lowercase hexadecimal digits.
 * @param pkg a pointer to package
 * @return a reference to an internal string
 */
const char *hbpm_pkg_get_md5sum(hbpm_pkg_t *pkg);

/** Returns the package's SHA256 checksum as a string.
 * The returned string is a sequence of 64 lowercase hexadecimal digits.
 * @param pkg a pointer to package
 * @return a reference to an internal string
 */
const char *hbpm_pkg_get_sha256sum(hbpm_pkg_t *pkg);

/** Returns the architecture for which the package was built.
 * @param pkg a pointer to package
 * @return a reference to an internal string
 */
const char *hbpm_pkg_get_arch(hbpm_pkg_t *pkg);

/** Returns the size of the package. This is only available for sync database
 * packages and package files, not those loaded from the local database.
 * @param pkg a pointer to package
 * @return the size of the package in bytes.
 */
off_t hbpm_pkg_get_size(hbpm_pkg_t *pkg);

/** Returns the installed size of the package.
 * @param pkg a pointer to package
 * @return the total size of files installed by the package.
 */
off_t hbpm_pkg_get_isize(hbpm_pkg_t *pkg);

/** Returns the package installation reason.
 * @param pkg a pointer to package
 * @return an enum member giving the install reason.
 */
hbpm_pkgreason_t hbpm_pkg_get_reason(hbpm_pkg_t *pkg);

/** Returns the list of package licenses.
 * @param pkg a pointer to package
 * @return a pointer to an internal list of strings.
 */
hbpm_list_t *hbpm_pkg_get_licenses(hbpm_pkg_t *pkg);

/** Returns the list of package groups.
 * @param pkg a pointer to package
 * @return a pointer to an internal list of strings.
 */
hbpm_list_t *hbpm_pkg_get_groups(hbpm_pkg_t *pkg);

/** Returns the list of package dependencies as hbpm_depend_t.
 * @param pkg a pointer to package
 * @return a reference to an internal list of hbpm_depend_t structures.
 */
hbpm_list_t *hbpm_pkg_get_depends(hbpm_pkg_t *pkg);

/** Returns the list of package optional dependencies.
 * @param pkg a pointer to package
 * @return a reference to an internal list of hbpm_depend_t structures.
 */
hbpm_list_t *hbpm_pkg_get_optdepends(hbpm_pkg_t *pkg);

/** Returns a list of package check dependencies
 * @param pkg a pointer to package
 * @return a reference to an internal list of hbpm_depend_t structures.
 */
hbpm_list_t *hbpm_pkg_get_checkdepends(hbpm_pkg_t *pkg);

/** Returns a list of package make dependencies
 * @param pkg a pointer to package
 * @return a reference to an internal list of hbpm_depend_t structures.
 */
hbpm_list_t *hbpm_pkg_get_makedepends(hbpm_pkg_t *pkg);

/** Returns the list of packages conflicting with pkg.
 * @param pkg a pointer to package
 * @return a reference to an internal list of hbpm_depend_t structures.
 */
hbpm_list_t *hbpm_pkg_get_conflicts(hbpm_pkg_t *pkg);

/** Returns the list of packages provided by pkg.
 * @param pkg a pointer to package
 * @return a reference to an internal list of hbpm_depend_t structures.
 */
hbpm_list_t *hbpm_pkg_get_provides(hbpm_pkg_t *pkg);

/** Returns the list of packages to be replaced by pkg.
 * @param pkg a pointer to package
 * @return a reference to an internal list of hbpm_depend_t structures.
 */
hbpm_list_t *hbpm_pkg_get_replaces(hbpm_pkg_t *pkg);

/** Returns the list of files installed by pkg.
 * The filenames are relative to the install root,
 * and do not include leading slashes.
 * @param pkg a pointer to package
 * @return a pointer to a filelist object containing a count and an array of
 * package file objects
 */
hbpm_filelist_t *hbpm_pkg_get_files(hbpm_pkg_t *pkg);

/** Returns the list of files backed up when installing pkg.
 * @param pkg a pointer to package
 * @return a reference to a list of hbpm_backup_t objects
 */
hbpm_list_t *hbpm_pkg_get_backup(hbpm_pkg_t *pkg);

/** Returns the database containing pkg.
 * Returns a pointer to the hbpm_db_t structure the package is
 * originating from, or NULL if the package was loaded from a file.
 * @param pkg a pointer to package
 * @return a pointer to the DB containing pkg, or NULL.
 */
hbpm_db_t *hbpm_pkg_get_db(hbpm_pkg_t *pkg);

/** Returns the base64 encoded package signature.
 * @param pkg a pointer to package
 * @return a reference to an internal string
 */
const char *hbpm_pkg_get_base64_sig(hbpm_pkg_t *pkg);

/** Returns the method used to validate a package during install.
 * @param pkg a pointer to package
 * @return an enum member giving the validation method
 */
int hbpm_pkg_get_validation(hbpm_pkg_t *pkg);

/* End of hbpm_pkg_t accessors */
/* @} */

/** Open a package changelog for reading.
 * Similar to fopen in functionality, except that the returned 'file
 * stream' could really be from an archive as well as from the database.
 * @param pkg the package to read the changelog of (either file or db)
 * @return a 'file stream' to the package changelog
 */
void *hbpm_pkg_changelog_open(hbpm_pkg_t *pkg);

/** Read data from an open changelog 'file stream'.
 * Similar to fread in functionality, this function takes a buffer and
 * amount of data to read. If an error occurs pm_errno will be set.
 * @param ptr a buffer to fill with raw changelog data
 * @param size the size of the buffer
 * @param pkg the package that the changelog is being read from
 * @param fp a 'file stream' to the package changelog
 * @return the number of characters read, or 0 if there is no more data or an
 * error occurred.
 */
size_t hbpm_pkg_changelog_read(void *ptr, size_t size,
		const hbpm_pkg_t *pkg, void *fp);

/** Close a package changelog for reading.
 * @param pkg the package to close the changelog of (either file or db)
 * @return 0 on success, -1 on error
 */
int hbpm_pkg_changelog_close(const hbpm_pkg_t *pkg, void *fp);

/** Open a package mtree file for reading.
 * @param pkg the local package to read the mtree of
 * @return an archive structure for the package mtree file
 */
struct archive *hbpm_pkg_mtree_open(hbpm_pkg_t *pkg);

/** Read next entry from a package mtree file.
 * @param pkg the package that the mtree file is being read from
 * @param archive the archive structure reading from the mtree file
 * @param entry an archive_entry to store the entry header information
 * @return 0 on success, 1 if end of archive is reached, -1 otherwise.
 */
int hbpm_pkg_mtree_next(const hbpm_pkg_t *pkg, struct archive *archive,
		struct archive_entry **entry);

/** Close a package mtree file.
 * @param pkg the local package to close the mtree of
 * @param the archive to close
 */
int hbpm_pkg_mtree_close(const hbpm_pkg_t *pkg, struct archive *archive);

/** Returns whether the package has an install scriptlet.
 * @return 0 if FALSE, TRUE otherwise
 */
int hbpm_pkg_has_scriptlet(hbpm_pkg_t *pkg);

/** Returns the size of the files that will be downloaded to install a
 * package.
 * @param newpkg the new package to upgrade to
 * @return the size of the download
 */
off_t hbpm_pkg_download_size(hbpm_pkg_t *newpkg);

/** Set install reason for a package in the local database.
 * The provided package object must be from the local database or this method
 * will fail. The write to the local database is performed immediately.
 * @param pkg the package to update
 * @param reason the new install reason
 * @return 0 on success, -1 on error (pm_errno is set accordingly)
 */
int hbpm_pkg_set_reason(hbpm_pkg_t *pkg, hbpm_pkgreason_t reason);


/* End of hbpm_pkg */
/** @} */

/*
 * Filelists
 */

/** Determines whether a package filelist contains a given path.
 * The provided path should be relative to the install root with no leading
 * slashes, e.g. "etc/localtime". When searching for directories, the path must
 * have a trailing slash.
 * @param filelist a pointer to a package filelist
 * @param path the path to search for in the package
 * @return a pointer to the matching file or NULL if not found
 */
hbpm_file_t *hbpm_filelist_contains(hbpm_filelist_t *filelist, const char *path);

/*
 * Signatures
 */

/**
 * Check the PGP signature for the given package file.
 * @param pkg the package to check
 * @param siglist a pointer to storage for signature results
 * @return a int value : 0 (valid), 1 (invalid), -1 (an error occurred)
 */
int hbpm_pkg_check_pgp_signature(hbpm_pkg_t *pkg, hbpm_siglist_t *siglist);

/**
 * Check the PGP signature for the given database.
 * @param db the database to check
 * @param siglist a pointer to storage for signature results
 * @return a int value : 0 (valid), 1 (invalid), -1 (an error occurred)
 */
int hbpm_db_check_pgp_signature(hbpm_db_t *db, hbpm_siglist_t *siglist);

/**
 * Clean up and free a signature result list.
 * Note that this does not free the siglist object itself in case that
 * was allocated on the stack; this is the responsibility of the caller.
 * @param siglist a pointer to storage for signature results
 * @return 0 on success, -1 on error
 */
int hbpm_siglist_cleanup(hbpm_siglist_t *siglist);

/**
 * Decode a loaded signature in base64 form.
 * @param base64_data the signature to attempt to decode
 * @param data the decoded data; must be freed by the caller
 * @param data_len the length of the returned data
 * @return 0 on success, -1 on failure to properly decode
 */
int hbpm_decode_signature(const char *base64_data,
		unsigned char **data, size_t *data_len);

/**
 * Extract the Issuer Key ID from a signature
 * @param sig PGP signature
 * @param len length of signature
 * @param keys a pointer to storage for key IDs
 * @return 0 on success, -1 on error
 */
int hbpm_extract_keyid(hbpm_handle_t *handle, const char *identifier,
		const unsigned char *sig, const size_t len, hbpm_list_t **keys);

/*
 * Groups
 */

/** Find group members across a list of databases.
 * If a member exists in several databases, only the first database is used.
 * IgnorePkg is also handled.
 * @param dbs the list of ahbpm_db_t *
 * @param name the name of the group
 * @return the list of ahbpm_pkg_t * (caller is responsible for ahbpm_list_free)
 */
hbpm_list_t *hbpm_find_group_pkgs(hbpm_list_t *dbs, const char *name);

/*
 * Sync
 */

/** Check for new version of pkg in sync repos
 * (only the first occurrence is considered in sync)
 */
hbpm_pkg_t *hbpm_sync_get_new_version(hbpm_pkg_t *pkg, hbpm_list_t *dbs_sync);

/** @addtogroup hbpm_api_trans Transaction Functions
 * Functions to manipulate libhbpm transactions
 * @{
 */

/** Transaction flags */
typedef enum _hbpm_transflag_t {
	/** Ignore dependency checks. */
	HBPM_TRANS_FLAG_NODEPS = 1,
	/* (1 << 1) flag can go here */
	/** Delete files even if they are tagged as backup. */
	HBPM_TRANS_FLAG_NOSAVE = (1 << 2),
	/** Ignore version numbers when checking dependencies. */
	HBPM_TRANS_FLAG_NODEPVERSION = (1 << 3),
	/** Remove also any packages depending on a package being removed. */
	HBPM_TRANS_FLAG_CASCADE = (1 << 4),
	/** Remove packages and their unneeded deps (not explicitly installed). */
	HBPM_TRANS_FLAG_RECURSE = (1 << 5),
	/** Modify database but do not commit changes to the filesystem. */
	HBPM_TRANS_FLAG_DBONLY = (1 << 6),
	/* (1 << 7) flag can go here */
	/** Use HBPM_PKG_REASON_DEPEND when installing packages. */
	HBPM_TRANS_FLAG_ALLDEPS = (1 << 8),
	/** Only download packages and do not actually install. */
	HBPM_TRANS_FLAG_DOWNLOADONLY = (1 << 9),
	/** Do not execute install scriptlets after installing. */
	HBPM_TRANS_FLAG_NOSCRIPTLET = (1 << 10),
	/** Ignore dependency conflicts. */
	HBPM_TRANS_FLAG_NOCONFLICTS = (1 << 11),
	/* (1 << 12) flag can go here */
	/** Do not install a package if it is already installed and up to date. */
	HBPM_TRANS_FLAG_NEEDED = (1 << 13),
	/** Use HBPM_PKG_REASON_EXPLICIT when installing packages. */
	HBPM_TRANS_FLAG_ALLEXPLICIT = (1 << 14),
	/** Do not remove a package if it is needed by another one. */
	HBPM_TRANS_FLAG_UNNEEDED = (1 << 15),
	/** Remove also explicitly installed unneeded deps (use with HBPM_TRANS_FLAG_RECURSE). */
	HBPM_TRANS_FLAG_RECURSEALL = (1 << 16),
	/** Do not lock the database during the operation. */
	HBPM_TRANS_FLAG_NOLOCK = (1 << 17)
} hbpm_transflag_t;

/** Returns the bitfield of flags for the current transaction.
 * @param handle the context handle
 * @return the bitfield of transaction flags
 */
int hbpm_trans_get_flags(hbpm_handle_t *handle);

/** Returns a list of packages added by the transaction.
 * @param handle the context handle
 * @return a list of hbpm_pkg_t structures
 */
hbpm_list_t *hbpm_trans_get_add(hbpm_handle_t *handle);

/** Returns the list of packages removed by the transaction.
 * @param handle the context handle
 * @return a list of hbpm_pkg_t structures
 */
hbpm_list_t *hbpm_trans_get_remove(hbpm_handle_t *handle);

/** Initialize the transaction.
 * @param handle the context handle
 * @param flags flags of the transaction (like nodeps, etc; see hbpm_transflag_t)
 * @return 0 on success, -1 on error (pm_errno is set accordingly)
 */
int hbpm_trans_init(hbpm_handle_t *handle, int flags);

/** Prepare a transaction.
 * @param handle the context handle
 * @param data the address of an hbpm_list where a list
 * of hbpm_depmissing_t objects is dumped (conflicting packages)
 * @return 0 on success, -1 on error (pm_errno is set accordingly)
 */
int hbpm_trans_prepare(hbpm_handle_t *handle, hbpm_list_t **data);

/** Commit a transaction.
 * @param handle the context handle
 * @param data the address of an hbpm_list where detailed description
 * of an error can be dumped (i.e. list of conflicting files)
 * @return 0 on success, -1 on error (pm_errno is set accordingly)
 */
int hbpm_trans_commit(hbpm_handle_t *handle, hbpm_list_t **data);

/** Interrupt a transaction.
 * @param handle the context handle
 * @return 0 on success, -1 on error (pm_errno is set accordingly)
 */
int hbpm_trans_interrupt(hbpm_handle_t *handle);

/** Release a transaction.
 * @param handle the context handle
 * @return 0 on success, -1 on error (pm_errno is set accordingly)
 */
int hbpm_trans_release(hbpm_handle_t *handle);
/** @} */

/** @name Common Transactions */
/** @{ */

/** Search for packages to upgrade and add them to the transaction.
 * @param handle the context handle
 * @param enable_downgrade allow downgrading of packages if the remote version is lower
 * @return 0 on success, -1 on error (pm_errno is set accordingly)
 */
int hbpm_sync_sysupgrade(hbpm_handle_t *handle, int enable_downgrade);

/** Add a package to the transaction.
 * If the package was loaded by hbpm_pkg_load(), it will be freed upon
 * hbpm_trans_release() invocation.
 * @param handle the context handle
 * @param pkg the package to add
 * @return 0 on success, -1 on error (pm_errno is set accordingly)
 */
int hbpm_add_pkg(hbpm_handle_t *handle, hbpm_pkg_t *pkg);

/** Add a package removal action to the transaction.
 * @param handle the context handle
 * @param pkg the package to uninstall
 * @return 0 on success, -1 on error (pm_errno is set accordingly)
 */
int hbpm_remove_pkg(hbpm_handle_t *handle, hbpm_pkg_t *pkg);

/** @} */

/** @addtogroup hbpm_api_depends Dependency Functions
 * Functions dealing with libhbpm representation of dependency
 * information.
 * @{
 */

/** Checks dependencies and returns missing ones in a list.
 * Dependencies can include versions with depmod operators.
 * @param handle the context handle
 * @param pkglist the list of local packages
 * @param remove an hbpm_list_t* of packages to be removed
 * @param upgrade an hbpm_list_t* of packages to be upgraded (remove-then-upgrade)
 * @param reversedeps handles the backward dependencies
 * @return an hbpm_list_t* of hbpm_depmissing_t pointers.
 */
hbpm_list_t *hbpm_checkdeps(hbpm_handle_t *handle, hbpm_list_t *pkglist,
		hbpm_list_t *remove, hbpm_list_t *upgrade, int reversedeps);

/** Find a package satisfying a specified dependency.
 * The dependency can include versions with depmod operators.
 * @param pkgs an hbpm_list_t* of hbpm_pkg_t where the satisfier will be searched
 * @param depstring package or provision name, versioned or not
 * @return a hbpm_pkg_t* satisfying depstring
 */
hbpm_pkg_t *hbpm_find_satisfier(hbpm_list_t *pkgs, const char *depstring);

/** Find a package satisfying a specified dependency.
 * First look for a literal, going through each db one by one. Then look for
 * providers. The first satisfier that belongs to an installed package is
 * returned. If no providers belong to an installed package then an
 * hbpm_question_select_provider_t is created to select the provider.
 * The dependency can include versions with depmod operators.
 *
 * @param handle the context handle
 * @param dbs an hbpm_list_t* of hbpm_db_t where the satisfier will be searched
 * @param depstring package or provision name, versioned or not
 * @return a hbpm_pkg_t* satisfying depstring
 */
hbpm_pkg_t *hbpm_find_dbs_satisfier(hbpm_handle_t *handle,
		hbpm_list_t *dbs, const char *depstring);

/**
 * @brief Check the package conflicts in a database
 *
 * @param handle the context handle
 * @param pkglist the list of packages to check
 *
 * @return an hbpm_list_t of hbpm_conflict_t
 */
hbpm_list_t *hbpm_checkconflicts(hbpm_handle_t *handle, hbpm_list_t *pkglist);

/** Returns a newly allocated string representing the dependency information.
 * @param dep a dependency info structure
 * @return a formatted string, e.g. "glibc>=2.12"
 */
char *hbpm_dep_compute_string(const hbpm_depend_t *dep);

/** Return a newly allocated dependency information parsed from a string
 * @param depstring a formatted string, e.g. "glibc=2.12"
 * @return a dependency info structure
 */
hbpm_depend_t *hbpm_dep_from_string(const char *depstring);

/** Free a dependency info structure
 * @param dep struct to free
 */
void hbpm_dep_free(hbpm_depend_t *dep);

/** @} */

/** @} */

/*
 * Helpers
 */

/* checksums */

/** \addtogroup hbpm_misc Miscellaneous Functions
 * @brief Various libhbpm functions
 * @{
 */

/** Get the md5 sum of file.
 * @param filename name of the file
 * @return the checksum on success, NULL on error
 */
char *hbpm_compute_md5sum(const char *filename);

/** Get the sha256 sum of file.
 * @param filename name of the file
 * @return the checksum on success, NULL on error
 */
char *hbpm_compute_sha256sum(const char *filename);

/** @} */

/** \addtogroup hbpm_interface Interface Functions
 * @brief Functions to initialize and release libhbpm
 * @{
 */

/** Initializes the library.
 * Creates handle, connects to database and creates lockfile.
 * This must be called before any other functions are called.
 * @param root the root path for all filesystem operations
 * @param dbpath the absolute path to the libhbpm database
 * @param err an optional variable to hold any error return codes
 * @return a context handle on success, NULL on error, err will be set if provided
 */
hbpm_handle_t *hbpm_initialize(const char *root, const char *dbpath,
		hbpm_errno_t *err);

/** Release the library.
 * Disconnects from the database, removes handle and lockfile
 * This should be the last hbpm call you make.
 * After this returns, handle should be considered invalid and cannot be reused
 * in any way.
 * @param myhandle the context handle
 * @return 0 on success, -1 on error
 */
int hbpm_release(hbpm_handle_t *handle);

/** @} */

/** Remove the database lock file
 * @param handle the context handle
 * @return 0 on success, -1 on error
 *
 * @note Safe to call from inside signal handlers.
 */
int hbpm_unlock(hbpm_handle_t *handle);

enum hbpm_caps {
	HBPM_CAPABILITY_NLS = (1 << 0),
	HBPM_CAPABILITY_DOWNLOADER = (1 << 1),
	HBPM_CAPABILITY_SIGNATURES = (1 << 2)
};

/** Get the version of library.
 * @return the library version, e.g. "6.0.4"
 * */
const char *hbpm_version(void);

/** Get the capabilities of the library.
 * @return a bitmask of the capabilities
 * */
int hbpm_capabilities(void);

/**
 * Free a fileconflict and its members.
 * @param conflict the fileconflict to free
 */
void hbpm_fileconflict_free(hbpm_fileconflict_t *conflict);
void hbpm_depmissing_free(hbpm_depmissing_t *miss);

/**
 * Free a conflict and its members.
 * @param conflict the conflict to free
 */
void hbpm_conflict_free(hbpm_conflict_t *conflict);

/* End of hbpm_api */
/** @} */

#ifdef __cplusplus
}
#endif
#endif /* HBPM_H */
