/*
 *  hbpm_list.h
 *
 *  Copyright (c) 2006-2019 Pacman Development Team <pacman-dev@archlinux.org>
 *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef HBPM_LIST_H
#define HBPM_LIST_H

#include <stdlib.h> /* size_t */

/* Note: hbpm_list.{c,h} are intended to be standalone files. Do not include
 * any other libhbpm headers.
 */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief Linked list type used by libhbpm.
 *
 * It is exposed so front ends can use it to prevent the need to reimplement
 * lists of their own; however, it is not required that the front end uses
 * it.
 */
typedef struct __hbpm_list_t {
	/** data held by the list node */
	void *data;
	/** pointer to the previous node */
	struct __hbpm_list_t *prev;
	/** pointer to the next node */
	struct __hbpm_list_t *next;
} hbpm_list_t;

#define FREELIST(p) do { hbpm_list_free_inner(p, free); hbpm_list_free(p); p = NULL; } while(0)

typedef void (*hbpm_list_fn_free)(void *); /* item deallocation callback */
typedef int (*hbpm_list_fn_cmp)(const void *, const void *); /* item comparison callback */

/* allocation */
void hbpm_list_free(hbpm_list_t *list);
void hbpm_list_free_inner(hbpm_list_t *list, hbpm_list_fn_free fn);

/* item mutators */
hbpm_list_t *hbpm_list_add(hbpm_list_t *list, void *data);
hbpm_list_t *hbpm_list_append(hbpm_list_t **list, void *data);
hbpm_list_t *hbpm_list_append_strdup(hbpm_list_t **list, const char *data);
hbpm_list_t *hbpm_list_add_sorted(hbpm_list_t *list, void *data, hbpm_list_fn_cmp fn);
hbpm_list_t *hbpm_list_join(hbpm_list_t *first, hbpm_list_t *second);
hbpm_list_t *hbpm_list_mmerge(hbpm_list_t *left, hbpm_list_t *right, hbpm_list_fn_cmp fn);
hbpm_list_t *hbpm_list_msort(hbpm_list_t *list, size_t n, hbpm_list_fn_cmp fn);
hbpm_list_t *hbpm_list_remove_item(hbpm_list_t *haystack, hbpm_list_t *item);
hbpm_list_t *hbpm_list_remove(hbpm_list_t *haystack, const void *needle, hbpm_list_fn_cmp fn, void **data);
hbpm_list_t *hbpm_list_remove_str(hbpm_list_t *haystack, const char *needle, char **data);
hbpm_list_t *hbpm_list_remove_dupes(const hbpm_list_t *list);
hbpm_list_t *hbpm_list_strdup(const hbpm_list_t *list);
hbpm_list_t *hbpm_list_copy(const hbpm_list_t *list);
hbpm_list_t *hbpm_list_copy_data(const hbpm_list_t *list, size_t size);
hbpm_list_t *hbpm_list_reverse(hbpm_list_t *list);

/* item accessors */
hbpm_list_t *hbpm_list_nth(const hbpm_list_t *list, size_t n);
hbpm_list_t *hbpm_list_next(const hbpm_list_t *list);
hbpm_list_t *hbpm_list_previous(const hbpm_list_t *list);
hbpm_list_t *hbpm_list_last(const hbpm_list_t *list);

/* misc */
size_t hbpm_list_count(const hbpm_list_t *list);
void *hbpm_list_find(const hbpm_list_t *haystack, const void *needle, hbpm_list_fn_cmp fn);
void *hbpm_list_find_ptr(const hbpm_list_t *haystack, const void *needle);
char *hbpm_list_find_str(const hbpm_list_t *haystack, const char *needle);
hbpm_list_t *hbpm_list_diff(const hbpm_list_t *lhs, const hbpm_list_t *rhs, hbpm_list_fn_cmp fn);
void hbpm_list_diff_sorted(const hbpm_list_t *left, const hbpm_list_t *right,
		hbpm_list_fn_cmp fn, hbpm_list_t **onlyleft, hbpm_list_t **onlyright);
void *hbpm_list_to_array(const hbpm_list_t *list, size_t n, size_t size);

#ifdef __cplusplus
}
#endif
#endif /* HBPM_LIST_H */
