/*
 *  package.h
 *
 *  Copyright (c) 2006-2019 Pacman Development Team <pacman-dev@archlinux.org>
 *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
 *  Copyright (c) 2005 by Aurelien Foret <orelien@chez.com>
 *  Copyright (c) 2006 by David Kimpe <dnaku@frugalware.org>
 *  Copyright (c) 2005, 2006 by Christian Hamar <krics@linuxforum.hu>
 *  Copyright (c) 2005, 2006 by Miklos Vajna <vmiklos@frugalware.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef HBPM_PACKAGE_H
#define HBPM_PACKAGE_H

#include <sys/types.h> /* off_t */

/* libarchive */
#include <archive.h>
#include <archive_entry.h>

#include "hbpm.h"
#include "backup.h"
#include "db.h"
#include "signing.h"

/** Package operations struct. This struct contains function pointers to
 * all methods used to access data in a package to allow for things such
 * as lazy package initialization (such as used by the file backend). Each
 * backend is free to define a struct containing pointers to a specific
 * implementation of these methods. Some backends may find using the
 * defined default_pkg_ops struct to work just fine for their needs.
 */
struct pkg_operations {
	const char *(*get_base) (hbpm_pkg_t *);
	const char *(*get_desc) (hbpm_pkg_t *);
	const char *(*get_url) (hbpm_pkg_t *);
	hbpm_time_t (*get_builddate) (hbpm_pkg_t *);
	hbpm_time_t (*get_installdate) (hbpm_pkg_t *);
	const char *(*get_packager) (hbpm_pkg_t *);
	const char *(*get_arch) (hbpm_pkg_t *);
	off_t (*get_isize) (hbpm_pkg_t *);
	hbpm_pkgreason_t (*get_reason) (hbpm_pkg_t *);
	int (*get_validation) (hbpm_pkg_t *);
	int (*has_scriptlet) (hbpm_pkg_t *);

	hbpm_list_t *(*get_licenses) (hbpm_pkg_t *);
	hbpm_list_t *(*get_groups) (hbpm_pkg_t *);
	hbpm_list_t *(*get_depends) (hbpm_pkg_t *);
	hbpm_list_t *(*get_optdepends) (hbpm_pkg_t *);
	hbpm_list_t *(*get_checkdepends) (hbpm_pkg_t *);
	hbpm_list_t *(*get_makedepends) (hbpm_pkg_t *);
	hbpm_list_t *(*get_conflicts) (hbpm_pkg_t *);
	hbpm_list_t *(*get_provides) (hbpm_pkg_t *);
	hbpm_list_t *(*get_replaces) (hbpm_pkg_t *);
	hbpm_filelist_t *(*get_files) (hbpm_pkg_t *);
	hbpm_list_t *(*get_backup) (hbpm_pkg_t *);

	void *(*changelog_open) (hbpm_pkg_t *);
	size_t (*changelog_read) (void *, size_t, const hbpm_pkg_t *, void *);
	int (*changelog_close) (const hbpm_pkg_t *, void *);

	struct archive *(*mtree_open) (hbpm_pkg_t *);
	int (*mtree_next) (const hbpm_pkg_t *, struct archive *, struct archive_entry **);
	int (*mtree_close) (const hbpm_pkg_t *, struct archive *);

	int (*force_load) (hbpm_pkg_t *);
};

/** The standard package operations struct. get fields directly from the
 * struct itself with no abstraction layer or any type of lazy loading.
 * The actual definition is in package.c so it can have access to the
 * default accessor functions which are defined there.
 */
extern struct pkg_operations default_pkg_ops;

struct __hbpm_pkg_t {
	unsigned long name_hash;
	char *filename;
	char *base;
	char *name;
	char *version;
	char *desc;
	char *url;
	char *packager;
	char *md5sum;
	char *sha256sum;
	char *base64_sig;
	char *arch;

	hbpm_time_t builddate;
	hbpm_time_t installdate;

	off_t size;
	off_t isize;
	off_t download_size;

	hbpm_handle_t *handle;

	hbpm_list_t *licenses;
	hbpm_list_t *replaces;
	hbpm_list_t *groups;
	hbpm_list_t *backup;
	hbpm_list_t *depends;
	hbpm_list_t *optdepends;
	hbpm_list_t *checkdepends;
	hbpm_list_t *makedepends;
	hbpm_list_t *conflicts;
	hbpm_list_t *provides;
	hbpm_list_t *removes; /* in transaction targets only */
	hbpm_pkg_t *oldpkg; /* in transaction targets only */

	struct pkg_operations *ops;

	hbpm_filelist_t files;

	/* origin == PKG_FROM_FILE, use pkg->origin_data.file
	 * origin == PKG_FROM_*DB, use pkg->origin_data.db */
	union {
		hbpm_db_t *db;
		char *file;
	} origin_data;

	hbpm_pkgfrom_t origin;
	hbpm_pkgreason_t reason;
	int scriptlet;

	/* Bitfield from hbpm_dbinfrq_t */
	int infolevel;
	/* Bitfield from hbpm_pkgvalidation_t */
	int validation;
};

hbpm_file_t *_hbpm_file_copy(hbpm_file_t *dest, const hbpm_file_t *src);

hbpm_pkg_t *_hbpm_pkg_new(void);
int _hbpm_pkg_dup(hbpm_pkg_t *pkg, hbpm_pkg_t **new_ptr);
void _hbpm_pkg_free(hbpm_pkg_t *pkg);
void _hbpm_pkg_free_trans(hbpm_pkg_t *pkg);

int _hbpm_pkg_validate_internal(hbpm_handle_t *handle,
		const char *pkgfile, hbpm_pkg_t *syncpkg, int level,
		hbpm_siglist_t **sigdata, int *validation);
hbpm_pkg_t *_hbpm_pkg_load_internal(hbpm_handle_t *handle,
		const char *pkgfile, int full);

int _hbpm_pkg_cmp(const void *p1, const void *p2);
int _hbpm_pkg_compare_versions(hbpm_pkg_t *local_pkg, hbpm_pkg_t *pkg);

#endif /* HBPM_PACKAGE_H */
