/*
 *  sync.c
 *
 *  Copyright (c) 2006-2019 Pacman Development Team <pacman-dev@archlinux.org>
 *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
 *  Copyright (c) 2005 by Aurelien Foret <orelien@chez.com>
 *  Copyright (c) 2005 by Christian Hamar <krics@linuxforum.hu>
 *  Copyright (c) 2005, 2006 by Miklos Vajna <vmiklos@frugalware.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/types.h> /* off_t */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h> /* intmax_t */
#include <unistd.h>
#include <limits.h>

/* libhbpm */
#include "sync.h"
#include "hbpm_list.h"
#include "log.h"
#include "package.h"
#include "db.h"
#include "deps.h"
#include "conflict.h"
#include "trans.h"
#include "add.h"
#include "util.h"
#include "handle.h"
#include "hbpm.h"
#include "dload.h"
#include "remove.h"
#include "diskspace.h"
#include "signing.h"

struct keyinfo_t {
       char* uid;
       char* keyid;
};

hbpm_pkg_t SYMEXPORT *hbpm_sync_get_new_version(hbpm_pkg_t *pkg, hbpm_list_t *dbs_sync)
{
	hbpm_list_t *i;
	hbpm_pkg_t *spkg = NULL;

	ASSERT(pkg != NULL, return NULL);
	pkg->handle->pm_errno = HBPM_ERR_OK;

	for(i = dbs_sync; !spkg && i; i = i->next) {
		hbpm_db_t *db = i->data;
		spkg = _hbpm_db_get_pkgfromcache(db, pkg->name);
	}

	if(spkg == NULL) {
		_hbpm_log(pkg->handle, HBPM_LOG_DEBUG, "'%s' not found in sync db => no upgrade\n",
				pkg->name);
		return NULL;
	}

	/* compare versions and see if spkg is an upgrade */
	if(_hbpm_pkg_compare_versions(spkg, pkg) > 0) {
		_hbpm_log(pkg->handle, HBPM_LOG_DEBUG, "new version of '%s' found (%s => %s)\n",
					pkg->name, pkg->version, spkg->version);
		return spkg;
	}
	/* spkg is not an upgrade */
	return NULL;
}

static int check_literal(hbpm_handle_t *handle, hbpm_pkg_t *lpkg,
		hbpm_pkg_t *spkg, int enable_downgrade)
{
	/* 1. literal was found in sdb */
	int cmp = _hbpm_pkg_compare_versions(spkg, lpkg);
	if(cmp > 0) {
		_hbpm_log(handle, HBPM_LOG_DEBUG, "new version of '%s' found (%s => %s)\n",
				lpkg->name, lpkg->version, spkg->version);
		/* check IgnorePkg/IgnoreGroup */
		if(hbpm_pkg_should_ignore(handle, spkg)
				|| hbpm_pkg_should_ignore(handle, lpkg)) {
			_hbpm_log(handle, HBPM_LOG_WARNING, _("%s: ignoring package upgrade (%s => %s)\n"),
					lpkg->name, lpkg->version, spkg->version);
		} else {
			_hbpm_log(handle, HBPM_LOG_DEBUG, "adding package %s-%s to the transaction targets\n",
					spkg->name, spkg->version);
			return 1;
		}
	} else if(cmp < 0) {
		if(enable_downgrade) {
			/* check IgnorePkg/IgnoreGroup */
			if(hbpm_pkg_should_ignore(handle, spkg)
					|| hbpm_pkg_should_ignore(handle, lpkg)) {
				_hbpm_log(handle, HBPM_LOG_WARNING, _("%s: ignoring package downgrade (%s => %s)\n"),
						lpkg->name, lpkg->version, spkg->version);
			} else {
				_hbpm_log(handle, HBPM_LOG_WARNING, _("%s: downgrading from version %s to version %s\n"),
						lpkg->name, lpkg->version, spkg->version);
				return 1;
			}
		} else {
			hbpm_db_t *sdb = hbpm_pkg_get_db(spkg);
			_hbpm_log(handle, HBPM_LOG_WARNING, _("%s: local (%s) is newer than %s (%s)\n"),
					lpkg->name, lpkg->version, sdb->treename, spkg->version);
		}
	}
	return 0;
}

static hbpm_list_t *check_replacers(hbpm_handle_t *handle, hbpm_pkg_t *lpkg,
		hbpm_db_t *sdb)
{
	/* 2. search for replacers in sdb */
	hbpm_list_t *replacers = NULL;
	hbpm_list_t *k;
	_hbpm_log(handle, HBPM_LOG_DEBUG,
			"searching for replacements for %s in %s\n",
			lpkg->name, sdb->treename);
	for(k = _hbpm_db_get_pkgcache(sdb); k; k = k->next) {
		int found = 0;
		hbpm_pkg_t *spkg = k->data;
		hbpm_list_t *l;
		for(l = hbpm_pkg_get_replaces(spkg); l; l = l->next) {
			hbpm_depend_t *replace = l->data;
			/* we only want to consider literal matches at this point. */
			if(_hbpm_depcmp_literal(lpkg, replace)) {
				found = 1;
				break;
			}
		}
		if(found) {
			hbpm_question_replace_t question = {
				.type = HBPM_QUESTION_REPLACE_PKG,
				.replace = 0,
				.oldpkg = lpkg,
				.newpkg = spkg,
				.newdb = sdb
			};
			hbpm_pkg_t *tpkg;
			/* check IgnorePkg/IgnoreGroup */
			if(hbpm_pkg_should_ignore(handle, spkg)
					|| hbpm_pkg_should_ignore(handle, lpkg)) {
				_hbpm_log(handle, HBPM_LOG_WARNING,
						_("ignoring package replacement (%s-%s => %s-%s)\n"),
						lpkg->name, lpkg->version, spkg->name, spkg->version);
				continue;
			}

			QUESTION(handle, &question);
			if(!question.replace) {
				continue;
			}

			/* If spkg is already in the target list, we append lpkg to spkg's
			 * removes list */
			tpkg = hbpm_pkg_find(handle->trans->add, spkg->name);
			if(tpkg) {
				/* sanity check, multiple repos can contain spkg->name */
				if(tpkg->origin_data.db != sdb) {
					_hbpm_log(handle, HBPM_LOG_WARNING, _("cannot replace %s by %s\n"),
							lpkg->name, spkg->name);
					continue;
				}
				_hbpm_log(handle, HBPM_LOG_DEBUG, "appending %s to the removes list of %s\n",
						lpkg->name, tpkg->name);
				tpkg->removes = hbpm_list_add(tpkg->removes, lpkg);
				/* check the to-be-replaced package's reason field */
				if(hbpm_pkg_get_reason(lpkg) == HBPM_PKG_REASON_EXPLICIT) {
					tpkg->reason = HBPM_PKG_REASON_EXPLICIT;
				}
			} else {
				/* add spkg to the target list */
				/* copy over reason */
				spkg->reason = hbpm_pkg_get_reason(lpkg);
				spkg->removes = hbpm_list_add(NULL, lpkg);
				_hbpm_log(handle, HBPM_LOG_DEBUG,
						"adding package %s-%s to the transaction targets\n",
						spkg->name, spkg->version);
				replacers = hbpm_list_add(replacers, spkg);
			}
		}
	}
	return replacers;
}

int SYMEXPORT hbpm_sync_sysupgrade(hbpm_handle_t *handle, int enable_downgrade)
{
	hbpm_list_t *i, *j;
	hbpm_trans_t *trans;

	CHECK_HANDLE(handle, return -1);
	trans = handle->trans;
	ASSERT(trans != NULL, RET_ERR(handle, HBPM_ERR_TRANS_NULL, -1));
	ASSERT(trans->state == STATE_INITIALIZED, RET_ERR(handle, HBPM_ERR_TRANS_NOT_INITIALIZED, -1));

	_hbpm_log(handle, HBPM_LOG_DEBUG, "checking for package upgrades\n");
	for(i = _hbpm_db_get_pkgcache(handle->db_local); i; i = i->next) {
		hbpm_pkg_t *lpkg = i->data;

		if(hbpm_pkg_find(trans->remove, lpkg->name)) {
			_hbpm_log(handle, HBPM_LOG_DEBUG, "%s is marked for removal -- skipping\n", lpkg->name);
			continue;
		}

		if(hbpm_pkg_find(trans->add, lpkg->name)) {
			_hbpm_log(handle, HBPM_LOG_DEBUG, "%s is already in the target list -- skipping\n", lpkg->name);
			continue;
		}

		/* Search for replacers then literal (if no replacer) in each sync database. */
		for(j = handle->dbs_sync; j; j = j->next) {
			hbpm_db_t *sdb = j->data;
			hbpm_list_t *replacers;

			if(!(sdb->usage & HBPM_DB_USAGE_UPGRADE)) {
				continue;
			}

			/* Check sdb */
			replacers = check_replacers(handle, lpkg, sdb);
			if(replacers) {
				trans->add = hbpm_list_join(trans->add, replacers);
				/* jump to next local package */
				break;
			} else {
				hbpm_pkg_t *spkg = _hbpm_db_get_pkgfromcache(sdb, lpkg->name);
				if(spkg) {
					if(check_literal(handle, lpkg, spkg, enable_downgrade)) {
						trans->add = hbpm_list_add(trans->add, spkg);
					}
					/* jump to next local package */
					break;
				}
			}
		}
	}

	return 0;
}

hbpm_list_t SYMEXPORT *hbpm_find_group_pkgs(hbpm_list_t *dbs,
		const char *name)
{
	hbpm_list_t *i, *j, *pkgs = NULL, *ignorelist = NULL;

	for(i = dbs; i; i = i->next) {
		hbpm_db_t *db = i->data;
		hbpm_group_t *grp = hbpm_db_get_group(db, name);

		if(!grp) {
			continue;
		}

		for(j = grp->packages; j; j = j->next) {
			hbpm_pkg_t *pkg = j->data;
			hbpm_trans_t *trans = db->handle->trans;

			if(hbpm_pkg_find(ignorelist, pkg->name)) {
				continue;
			}
			if(trans != NULL && trans->flags & HBPM_TRANS_FLAG_NEEDED) {
				hbpm_pkg_t *local = _hbpm_db_get_pkgfromcache(db->handle->db_local, pkg->name);
				if(local && _hbpm_pkg_compare_versions(pkg, local) == 0) {
					/* with the NEEDED flag, packages up to date are not reinstalled */
					_hbpm_log(db->handle, HBPM_LOG_WARNING, _("%s-%s is up to date -- skipping\n"),
							local->name, local->version);
					ignorelist = hbpm_list_add(ignorelist, pkg);
					continue;
				}
			}
			if(hbpm_pkg_should_ignore(db->handle, pkg)) {
				hbpm_question_install_ignorepkg_t question = {
					.type = HBPM_QUESTION_INSTALL_IGNOREPKG,
					.install = 0,
					.pkg = pkg
				};
				ignorelist = hbpm_list_add(ignorelist, pkg);
				QUESTION(db->handle, &question);
				if(!question.install) {
					continue;
				}
			}
			if(!hbpm_pkg_find(pkgs, pkg->name)) {
				pkgs = hbpm_list_add(pkgs, pkg);
			}
		}
	}
	hbpm_list_free(ignorelist);
	return pkgs;
}

/** Compute the size of the files that will be downloaded to install a
 * package.
 * @param newpkg the new package to upgrade to
 */
static int compute_download_size(hbpm_pkg_t *newpkg)
{
	const char *fname;
	char *fpath, *fnamepart = NULL;
	off_t size = 0;
	hbpm_handle_t *handle = newpkg->handle;
	int ret = 0;

	if(newpkg->origin != HBPM_PKG_FROM_SYNCDB) {
		newpkg->infolevel |= INFRQ_DSIZE;
		newpkg->download_size = 0;
		return 0;
	}

	ASSERT(newpkg->filename != NULL, RET_ERR(handle, HBPM_ERR_PKG_INVALID_NAME, -1));
	fname = newpkg->filename;
	fpath = _hbpm_filecache_find(handle, fname);

	/* downloaded file exists, so there's nothing to grab */
	if(fpath) {
		size = 0;
		goto finish;
	}

	CALLOC(fnamepart, strlen(fname) + 6, sizeof(char), return -1);
	sprintf(fnamepart, "%s.part", fname);
	fpath = _hbpm_filecache_find(handle, fnamepart);
	if(fpath) {
		struct stat st;
		if(stat(fpath, &st) == 0) {
			/* subtract the size of the .part file */
			_hbpm_log(handle, HBPM_LOG_DEBUG, "using (package - .part) size\n");
			size = newpkg->size - st.st_size;
			size = size < 0 ? 0 : size;
		}

		/* tell the caller that we have a partial */
		ret = 1;
	} else {
		size = newpkg->size;
	}

finish:
	_hbpm_log(handle, HBPM_LOG_DEBUG, "setting download size %jd for pkg %s\n",
			(intmax_t)size, newpkg->name);

	newpkg->infolevel |= INFRQ_DSIZE;
	newpkg->download_size = size;

	FREE(fpath);
	FREE(fnamepart);

	return ret;
}

int _hbpm_sync_prepare(hbpm_handle_t *handle, hbpm_list_t **data)
{
	hbpm_list_t *i, *j;
	hbpm_list_t *deps = NULL;
	hbpm_list_t *unresolvable = NULL;
	int from_sync = 0;
	int ret = 0;
	hbpm_trans_t *trans = handle->trans;
	hbpm_event_t event;

	if(data) {
		*data = NULL;
	}

	for(i = trans->add; i; i = i->next) {
		hbpm_pkg_t *spkg = i->data;
		if (spkg->origin == HBPM_PKG_FROM_SYNCDB){
			from_sync = 1;
			break;
		}
	}

	/* ensure all sync database are valid if we will be using them */
	for(i = handle->dbs_sync; i; i = i->next) {
		const hbpm_db_t *db = i->data;
		if(db->status & DB_STATUS_INVALID) {
			RET_ERR(handle, HBPM_ERR_DB_INVALID, -1);
		}
		/* missing databases are not allowed if we have sync targets */
		if(from_sync && db->status & DB_STATUS_MISSING) {
			RET_ERR(handle, HBPM_ERR_DB_NOT_FOUND, -1);
		}
	}

	if(!(trans->flags & HBPM_TRANS_FLAG_NODEPS)) {
		hbpm_list_t *resolved = NULL;
		hbpm_list_t *remove = hbpm_list_copy(trans->remove);
		hbpm_list_t *localpkgs;

		/* Build up list by repeatedly resolving each transaction package */
		/* Resolve targets dependencies */
		event.type = HBPM_EVENT_RESOLVEDEPS_START;
		EVENT(handle, &event);
		_hbpm_log(handle, HBPM_LOG_DEBUG, "resolving target's dependencies\n");

		/* build remove list for resolvedeps */
		for(i = trans->add; i; i = i->next) {
			hbpm_pkg_t *spkg = i->data;
			for(j = spkg->removes; j; j = j->next) {
				remove = hbpm_list_add(remove, j->data);
			}
		}

		/* Compute the fake local database for resolvedeps (partial fix for the
		 * phonon/qt issue) */
		localpkgs = hbpm_list_diff(_hbpm_db_get_pkgcache(handle->db_local),
				trans->add, _hbpm_pkg_cmp);

		/* Resolve packages in the transaction one at a time, in addition
		   building up a list of packages which could not be resolved. */
		for(i = trans->add; i; i = i->next) {
			hbpm_pkg_t *pkg = i->data;
			if(_hbpm_resolvedeps(handle, localpkgs, pkg, trans->add,
						&resolved, remove, data) == -1) {
				unresolvable = hbpm_list_add(unresolvable, pkg);
			}
			/* Else, [resolved] now additionally contains [pkg] and all of its
			   dependencies not already on the list */
		}
		hbpm_list_free(localpkgs);
		hbpm_list_free(remove);

		/* If there were unresolvable top-level packages, prompt the user to
		   see if they'd like to ignore them rather than failing the sync */
		if(unresolvable != NULL) {
			hbpm_question_remove_pkgs_t question = {
				.type = HBPM_QUESTION_REMOVE_PKGS,
				.skip = 0,
				.packages = unresolvable
			};
			QUESTION(handle, &question);
			if(question.skip) {
				/* User wants to remove the unresolvable packages from the
				   transaction. The packages will be removed from the actual
				   transaction when the transaction packages are replaced with a
				   dependency-reordered list below */
				handle->pm_errno = HBPM_ERR_OK;
				if(data) {
					hbpm_list_free_inner(*data,
							(hbpm_list_fn_free)hbpm_depmissing_free);
					hbpm_list_free(*data);
					*data = NULL;
				}
			} else {
				/* pm_errno was set by resolvedeps, callback may have overwrote it */
				hbpm_list_free(resolved);
				hbpm_list_free(unresolvable);
				ret = -1;
				GOTO_ERR(handle, HBPM_ERR_UNSATISFIED_DEPS, cleanup);
			}
		}

		/* Set DEPEND reason for pulled packages */
		for(i = resolved; i; i = i->next) {
			hbpm_pkg_t *pkg = i->data;
			if(!hbpm_pkg_find(trans->add, pkg->name)) {
				pkg->reason = HBPM_PKG_REASON_DEPEND;
			}
		}

		/* Unresolvable packages will be removed from the target list; set these
		 * aside in the transaction as a list we won't operate on. If we free them
		 * before the end of the transaction, we may kill pointers the frontend
		 * holds to package objects. */
		trans->unresolvable = unresolvable;

		hbpm_list_free(trans->add);
		trans->add = resolved;

		event.type = HBPM_EVENT_RESOLVEDEPS_DONE;
		EVENT(handle, &event);
	}

	if(!(trans->flags & HBPM_TRANS_FLAG_NOCONFLICTS)) {
		/* check for inter-conflicts and whatnot */
		event.type = HBPM_EVENT_INTERCONFLICTS_START;
		EVENT(handle, &event);

		_hbpm_log(handle, HBPM_LOG_DEBUG, "looking for conflicts\n");

		/* 1. check for conflicts in the target list */
		_hbpm_log(handle, HBPM_LOG_DEBUG, "check targets vs targets\n");
		deps = _hbpm_innerconflicts(handle, trans->add);

		for(i = deps; i; i = i->next) {
			hbpm_conflict_t *conflict = i->data;
			hbpm_pkg_t *rsync, *sync, *sync1, *sync2;

			/* have we already removed one of the conflicting targets? */
			sync1 = hbpm_pkg_find(trans->add, conflict->package1);
			sync2 = hbpm_pkg_find(trans->add, conflict->package2);
			if(!sync1 || !sync2) {
				continue;
			}

			_hbpm_log(handle, HBPM_LOG_DEBUG, "conflicting packages in the sync list: '%s' <-> '%s'\n",
					conflict->package1, conflict->package2);

			/* if sync1 provides sync2, we remove sync2 from the targets, and vice versa */
			hbpm_depend_t *dep1 = hbpm_dep_from_string(conflict->package1);
			hbpm_depend_t *dep2 = hbpm_dep_from_string(conflict->package2);
			if(_hbpm_depcmp(sync1, dep2)) {
				rsync = sync2;
				sync = sync1;
			} else if(_hbpm_depcmp(sync2, dep1)) {
				rsync = sync1;
				sync = sync2;
			} else {
				_hbpm_log(handle, HBPM_LOG_ERROR, _("unresolvable package conflicts detected\n"));
				handle->pm_errno = HBPM_ERR_CONFLICTING_DEPS;
				ret = -1;
				if(data) {
					hbpm_conflict_t *newconflict = _hbpm_conflict_dup(conflict);
					if(newconflict) {
						*data = hbpm_list_add(*data, newconflict);
					}
				}
				hbpm_list_free_inner(deps, (hbpm_list_fn_free)hbpm_conflict_free);
				hbpm_list_free(deps);
				hbpm_dep_free(dep1);
				hbpm_dep_free(dep2);
				goto cleanup;
			}
			hbpm_dep_free(dep1);
			hbpm_dep_free(dep2);

			/* Prints warning */
			_hbpm_log(handle, HBPM_LOG_WARNING,
					_("removing '%s' from target list because it conflicts with '%s'\n"),
					rsync->name, sync->name);
			trans->add = hbpm_list_remove(trans->add, rsync, _hbpm_pkg_cmp, NULL);
			/* rsync is not a transaction target anymore */
			trans->unresolvable = hbpm_list_add(trans->unresolvable, rsync);
		}

		hbpm_list_free_inner(deps, (hbpm_list_fn_free)hbpm_conflict_free);
		hbpm_list_free(deps);
		deps = NULL;

		/* 2. we check for target vs db conflicts (and resolve)*/
		_hbpm_log(handle, HBPM_LOG_DEBUG, "check targets vs db and db vs targets\n");
		deps = _hbpm_outerconflicts(handle->db_local, trans->add);

		for(i = deps; i; i = i->next) {
			hbpm_question_conflict_t question = {
				.type = HBPM_QUESTION_CONFLICT_PKG,
				.remove = 0,
				.conflict = i->data
			};
			hbpm_conflict_t *conflict = i->data;
			int found = 0;

			/* if conflict->package2 (the local package) is not elected for removal,
			   we ask the user */
			if(hbpm_pkg_find(trans->remove, conflict->package2)) {
				found = 1;
			}
			for(j = trans->add; j && !found; j = j->next) {
				hbpm_pkg_t *spkg = j->data;
				if(hbpm_pkg_find(spkg->removes, conflict->package2)) {
					found = 1;
				}
			}
			if(found) {
				continue;
			}

			_hbpm_log(handle, HBPM_LOG_DEBUG, "package '%s' conflicts with '%s'\n",
					conflict->package1, conflict->package2);

			QUESTION(handle, &question);
			if(question.remove) {
				/* append to the removes list */
				hbpm_pkg_t *sync = hbpm_pkg_find(trans->add, conflict->package1);
				hbpm_pkg_t *local = _hbpm_db_get_pkgfromcache(handle->db_local, conflict->package2);
				_hbpm_log(handle, HBPM_LOG_DEBUG, "electing '%s' for removal\n", conflict->package2);
				sync->removes = hbpm_list_add(sync->removes, local);
			} else { /* abort */
				_hbpm_log(handle, HBPM_LOG_ERROR, _("unresolvable package conflicts detected\n"));
				handle->pm_errno = HBPM_ERR_CONFLICTING_DEPS;
				ret = -1;
				if(data) {
					hbpm_conflict_t *newconflict = _hbpm_conflict_dup(conflict);
					if(newconflict) {
						*data = hbpm_list_add(*data, newconflict);
					}
				}
				hbpm_list_free_inner(deps, (hbpm_list_fn_free)hbpm_conflict_free);
				hbpm_list_free(deps);
				goto cleanup;
			}
		}
		event.type = HBPM_EVENT_INTERCONFLICTS_DONE;
		EVENT(handle, &event);
		hbpm_list_free_inner(deps, (hbpm_list_fn_free)hbpm_conflict_free);
		hbpm_list_free(deps);
	}

	/* Build trans->remove list */
	for(i = trans->add; i; i = i->next) {
		hbpm_pkg_t *spkg = i->data;
		for(j = spkg->removes; j; j = j->next) {
			hbpm_pkg_t *rpkg = j->data;
			if(!hbpm_pkg_find(trans->remove, rpkg->name)) {
				hbpm_pkg_t *copy;
				_hbpm_log(handle, HBPM_LOG_DEBUG, "adding '%s' to remove list\n", rpkg->name);
				if(_hbpm_pkg_dup(rpkg, &copy) == -1) {
					return -1;
				}
				trans->remove = hbpm_list_add(trans->remove, copy);
			}
		}
	}

	if(!(trans->flags & HBPM_TRANS_FLAG_NODEPS)) {
		_hbpm_log(handle, HBPM_LOG_DEBUG, "checking dependencies\n");
		deps = hbpm_checkdeps(handle, _hbpm_db_get_pkgcache(handle->db_local),
				trans->remove, trans->add, 1);
		if(deps) {
			handle->pm_errno = HBPM_ERR_UNSATISFIED_DEPS;
			ret = -1;
			if(data) {
				*data = deps;
			} else {
				hbpm_list_free_inner(deps,
						(hbpm_list_fn_free)hbpm_depmissing_free);
				hbpm_list_free(deps);
			}
			goto cleanup;
		}
	}
	for(i = trans->add; i; i = i->next) {
		/* update download size field */
		hbpm_pkg_t *spkg = i->data;
		hbpm_pkg_t *lpkg = hbpm_db_get_pkg(handle->db_local, spkg->name);
		if(compute_download_size(spkg) < 0) {
			ret = -1;
			goto cleanup;
		}
		if(lpkg && _hbpm_pkg_dup(lpkg, &spkg->oldpkg) != 0) {
			ret = -1;
			goto cleanup;
		}
	}

cleanup:
	return ret;
}

off_t SYMEXPORT hbpm_pkg_download_size(hbpm_pkg_t *newpkg)
{
	if(!(newpkg->infolevel & INFRQ_DSIZE)) {
		compute_download_size(newpkg);
	}
	return newpkg->download_size;
}

/**
 * Prompts to delete the file now that we know it is invalid.
 * @param handle the context handle
 * @param filename the absolute path of the file to test
 * @param reason an error code indicating the reason for package invalidity
 *
 * @return 1 if file was removed, 0 otherwise
 */
static int prompt_to_delete(hbpm_handle_t *handle, const char *filepath,
		hbpm_errno_t reason)
{
	hbpm_question_corrupted_t question = {
		.type = HBPM_QUESTION_CORRUPTED_PKG,
		.remove = 0,
		.filepath = filepath,
		.reason = reason
	};
	QUESTION(handle, &question);
	if(question.remove) {
		unlink(filepath);
	}
	return question.remove;
}

static int find_dl_candidates(hbpm_handle_t *handle, hbpm_list_t **files)
{
	for(hbpm_list_t *i = handle->trans->add; i; i = i->next) {
		hbpm_pkg_t *spkg = i->data;

		if(spkg->origin != HBPM_PKG_FROM_FILE) {
			char *fpath = NULL;
			hbpm_db_t *repo = spkg->origin_data.db;

			if(!repo->servers) {
				handle->pm_errno = HBPM_ERR_SERVER_NONE;
				_hbpm_log(handle, HBPM_LOG_ERROR, "%s: %s\n",
						hbpm_strerror(handle->pm_errno), repo->treename);
				return 1;
			}

			ASSERT(spkg->filename != NULL, RET_ERR(handle, HBPM_ERR_PKG_INVALID_NAME, -1));

			if(spkg->download_size == 0) {
				/* check for file in cache - allows us to handle complete .part files */
				fpath = _hbpm_filecache_find(handle, spkg->filename);
			}

			if(spkg->download_size != 0 || !fpath) {
				*files = hbpm_list_add(*files, spkg);
			}

			FREE(fpath);
		}
	}

	return 0;
}

static int download_single_file(hbpm_handle_t *handle, struct dload_payload *payload,
		const char *cachedir)
{
	hbpm_event_pkgdownload_t event = {
		.type = HBPM_EVENT_PKGDOWNLOAD_START,
		.file = payload->remote_name
	};
	const hbpm_list_t *server;

	payload->handle = handle;
	payload->allow_resume = 1;

	EVENT(handle, &event);
	for(server = payload->servers; server; server = server->next) {
		const char *server_url = server->data;
		size_t len;

		/* print server + filename into a buffer */
		len = strlen(server_url) + strlen(payload->remote_name) + 2;
		MALLOC(payload->fileurl, len, RET_ERR(handle, HBPM_ERR_MEMORY, -1));
		snprintf(payload->fileurl, len, "%s/%s", server_url, payload->remote_name);

		if(_hbpm_download(payload, cachedir, NULL, NULL) != -1) {
			event.type = HBPM_EVENT_PKGDOWNLOAD_DONE;
			EVENT(handle, &event);
			return 0;
		}
		_hbpm_dload_payload_reset_for_retry(payload);
	}

	event.type = HBPM_EVENT_PKGDOWNLOAD_FAILED;
	EVENT(handle, &event);
	return -1;
}

static int download_files(hbpm_handle_t *handle)
{
	const char *cachedir;
	hbpm_list_t *i, *files = NULL;
	int errors = 0;
	hbpm_event_t event;

	cachedir = _hbpm_filecache_setup(handle);
	handle->trans->state = STATE_DOWNLOADING;

	/* Total progress - figure out the total download size if required to
	 * pass to the callback. This function is called once, and it is up to the
	 * frontend to compute incremental progress. */
	if(handle->totaldlcb) {
		off_t total_size = (off_t)0;
		/* sum up the download size for each package and store total */
		for(i = handle->trans->add; i; i = i->next) {
			hbpm_pkg_t *spkg = i->data;
			total_size += spkg->download_size;
		}
		handle->totaldlcb(total_size);
	}

	errors += find_dl_candidates(handle, &files);

	if(files && !errors) {
		/* check for necessary disk space for download */
		if(handle->checkspace) {
			off_t *file_sizes;
			size_t idx, num_files;
			int ret;

			_hbpm_log(handle, HBPM_LOG_DEBUG, "checking available disk space for download\n");

			num_files = hbpm_list_count(files);
			CALLOC(file_sizes, num_files, sizeof(off_t), goto finish);

			for(i = files, idx = 0; i; i = i->next, idx++) {
				const hbpm_pkg_t *pkg = i->data;
				file_sizes[idx] = pkg->size;
			}

			ret = _hbpm_check_downloadspace(handle, cachedir, num_files, file_sizes);
			free(file_sizes);

			if(ret != 0) {
				errors++;
				goto finish;
			}
		}

		event.type = HBPM_EVENT_RETRIEVE_START;
		EVENT(handle, &event);
		event.type = HBPM_EVENT_RETRIEVE_DONE;
		for(i = files; i; i = i->next) {
			const hbpm_pkg_t *pkg = i->data;
			struct dload_payload payload = {0};

			STRDUP(payload.remote_name, pkg->filename, GOTO_ERR(handle, HBPM_ERR_MEMORY, finish));
			payload.servers = pkg->origin_data.db->servers;
			payload.max_size = pkg->size;

			if(download_single_file(handle, &payload, cachedir) == -1) {
				errors++;
				event.type = HBPM_EVENT_RETRIEVE_FAILED;
				_hbpm_log(handle, HBPM_LOG_WARNING, _("failed to retrieve some files\n"));
			}
			_hbpm_dload_payload_reset(&payload);
		}
		EVENT(handle, &event);
	}

finish:
	if(files) {
		hbpm_list_free(files);
	}

	for(i = handle->trans->add; i; i = i->next) {
		hbpm_pkg_t *pkg = i->data;
		pkg->infolevel &= ~INFRQ_DSIZE;
		pkg->download_size = 0;
	}

	/* clear out value to let callback know we are done */
	if(handle->totaldlcb) {
		handle->totaldlcb(0);
	}

	return errors;
}

#ifdef HAVE_LIBGPGME

static int key_cmp(const void *k1, const void *k2) {
	const struct keyinfo_t *key1 = k1;
	const char *key2 = k2;

	return strcmp(key1->keyid, key2);
}

static int check_keyring(hbpm_handle_t *handle)
{
	size_t current = 0, numtargs;
	hbpm_list_t *i, *errors = NULL;
	hbpm_event_t event;
	struct keyinfo_t *keyinfo;

	event.type = HBPM_EVENT_KEYRING_START;
	EVENT(handle, &event);

	numtargs = hbpm_list_count(handle->trans->add);

	for(i = handle->trans->add; i; i = i->next, current++) {
		hbpm_pkg_t *pkg = i->data;
		int level;

		int percent = (current * 100) / numtargs;
		PROGRESS(handle, HBPM_PROGRESS_KEYRING_START, "", percent,
				numtargs, current);

		if(pkg->origin == HBPM_PKG_FROM_FILE) {
			continue; /* pkg_load() has been already called, this package is valid */
		}

		level = hbpm_db_get_siglevel(hbpm_pkg_get_db(pkg));
		if((level & HBPM_SIG_PACKAGE) && pkg->base64_sig) {
			unsigned char *decoded_sigdata = NULL;
			size_t data_len;
			int decode_ret = hbpm_decode_signature(pkg->base64_sig,
					&decoded_sigdata, &data_len);
			if(decode_ret == 0) {
				hbpm_list_t *keys = NULL;
				if(hbpm_extract_keyid(handle, pkg->name, decoded_sigdata,
							data_len, &keys) == 0) {
					hbpm_list_t *k;
					for(k = keys; k; k = k->next) {
						char *key = k->data;
						if(!hbpm_list_find(errors, key, key_cmp) &&
								_hbpm_key_in_keychain(handle, key) == 0) {
							keyinfo = malloc(sizeof(struct keyinfo_t));
							if(!keyinfo) {
								break;
							}
							keyinfo->uid = strdup(pkg->packager);
							keyinfo->keyid = strdup(key);
							errors = hbpm_list_add(errors, keyinfo);
						}
					}
					FREELIST(keys);
				}
				free(decoded_sigdata);
			}
		}
	}

	PROGRESS(handle, HBPM_PROGRESS_KEYRING_START, "", 100,
			numtargs, current);
	event.type = HBPM_EVENT_KEYRING_DONE;
	EVENT(handle, &event);

	if(errors) {
		event.type = HBPM_EVENT_KEY_DOWNLOAD_START;
		EVENT(handle, &event);
		int fail = 0;
		hbpm_list_t *k;
		for(k = errors; k; k = k->next) {
			keyinfo = k->data;
			if(_hbpm_key_import(handle, keyinfo->uid, keyinfo->keyid) == -1) {
				fail = 1;
			}
			free(keyinfo->uid);
			free(keyinfo->keyid);

		}
		event.type = HBPM_EVENT_KEY_DOWNLOAD_DONE;
		EVENT(handle, &event);
		if(fail) {
			_hbpm_log(handle, HBPM_LOG_ERROR, _("required key missing from keyring\n"));
			return -1;
		}
	}

	return 0;
}
#endif /* HAVE_LIBGPGME */

static int check_validity(hbpm_handle_t *handle,
		size_t total, uint64_t total_bytes)
{
	struct validity {
		hbpm_pkg_t *pkg;
		char *path;
		hbpm_siglist_t *siglist;
		int siglevel;
		int validation;
		hbpm_errno_t error;
	};
	size_t current = 0;
	uint64_t current_bytes = 0;
	hbpm_list_t *i, *errors = NULL;
	hbpm_event_t event;

	/* Check integrity of packages */
	event.type = HBPM_EVENT_INTEGRITY_START;
	EVENT(handle, &event);

	for(i = handle->trans->add; i; i = i->next, current++) {
		struct validity v = { i->data, NULL, NULL, 0, 0, 0 };
		int percent = (int)(((double)current_bytes / total_bytes) * 100);

		PROGRESS(handle, HBPM_PROGRESS_INTEGRITY_START, "", percent,
				total, current);
		if(v.pkg->origin == HBPM_PKG_FROM_FILE) {
			continue; /* pkg_load() has been already called, this package is valid */
		}

		current_bytes += v.pkg->size;
		v.path = _hbpm_filecache_find(handle, v.pkg->filename);
		v.siglevel = hbpm_db_get_siglevel(hbpm_pkg_get_db(v.pkg));

		if(_hbpm_pkg_validate_internal(handle, v.path, v.pkg,
					v.siglevel, &v.siglist, &v.validation) == -1) {
			struct validity *invalid;
			v.error = handle->pm_errno;
			MALLOC(invalid, sizeof(struct validity), return -1);
			memcpy(invalid, &v, sizeof(struct validity));
			errors = hbpm_list_add(errors, invalid);
		} else {
			hbpm_siglist_cleanup(v.siglist);
			free(v.siglist);
			free(v.path);
			v.pkg->validation = v.validation;
		}
	}

	PROGRESS(handle, HBPM_PROGRESS_INTEGRITY_START, "", 100,
			total, current);
	event.type = HBPM_EVENT_INTEGRITY_DONE;
	EVENT(handle, &event);

	if(errors) {
		for(i = errors; i; i = i->next) {
			struct validity *v = i->data;
			switch(v->error) {
				case HBPM_ERR_PKG_MISSING_SIG:
					_hbpm_log(handle, HBPM_LOG_ERROR,
							_("%s: missing required signature\n"), v->pkg->name);
					break;
				case HBPM_ERR_PKG_INVALID_SIG:
					_hbpm_process_siglist(handle, v->pkg->name, v->siglist,
							v->siglevel & HBPM_SIG_PACKAGE_OPTIONAL,
							v->siglevel & HBPM_SIG_PACKAGE_MARGINAL_OK,
							v->siglevel & HBPM_SIG_PACKAGE_UNKNOWN_OK);
					/* fallthrough */
				case HBPM_ERR_PKG_INVALID_CHECKSUM:
					prompt_to_delete(handle, v->path, v->error);
					break;
				case HBPM_ERR_PKG_NOT_FOUND:
				case HBPM_ERR_BADPERMS:
				case HBPM_ERR_PKG_OPEN:
					_hbpm_log(handle, HBPM_LOG_ERROR, _("failed to read file %s: %s\n"), v->path, hbpm_strerror(v->error));
					break;
				default:
					/* ignore */
					break;
			}
			hbpm_siglist_cleanup(v->siglist);
			free(v->siglist);
			free(v->path);
			free(v);
		}
		hbpm_list_free(errors);

		if(handle->pm_errno == HBPM_ERR_OK) {
			RET_ERR(handle, HBPM_ERR_PKG_INVALID, -1);
		}
		return -1;
	}

	return 0;
}

static int load_packages(hbpm_handle_t *handle, hbpm_list_t **data,
		size_t total, size_t total_bytes)
{
	size_t current = 0, current_bytes = 0;
	int errors = 0;
	hbpm_list_t *i;
	hbpm_event_t event;

	/* load packages from disk now that they are known-valid */
	event.type = HBPM_EVENT_LOAD_START;
	EVENT(handle, &event);

	for(i = handle->trans->add; i; i = i->next, current++) {
		int error = 0;
		hbpm_pkg_t *spkg = i->data;
		char *filepath;
		int percent = (int)(((double)current_bytes / total_bytes) * 100);

		PROGRESS(handle, HBPM_PROGRESS_LOAD_START, "", percent,
				total, current);
		if(spkg->origin == HBPM_PKG_FROM_FILE) {
			continue; /* pkg_load() has been already called, this package is valid */
		}

		current_bytes += spkg->size;
		filepath = _hbpm_filecache_find(handle, spkg->filename);

		/* load the package file and replace pkgcache entry with it in the target list */
		/* TODO: hbpm_pkg_get_db() will not work on this target anymore */
		_hbpm_log(handle, HBPM_LOG_DEBUG,
				"replacing pkgcache entry with package file for target %s\n",
				spkg->name);
		hbpm_pkg_t *pkgfile =_hbpm_pkg_load_internal(handle, filepath, 1);
		if(!pkgfile) {
			_hbpm_log(handle, HBPM_LOG_DEBUG, "failed to load pkgfile internal\n");
			error = 1;
		} else {
			if(strcmp(spkg->name, pkgfile->name) != 0) {
				_hbpm_log(handle, HBPM_LOG_DEBUG,
						"internal package name mismatch, expected: '%s', actual: '%s'\n",
						spkg->name, pkgfile->name);
				error = 1;
			}
			if(strcmp(spkg->version, pkgfile->version) != 0) {
				_hbpm_log(handle, HBPM_LOG_DEBUG,
						"internal package version mismatch, expected: '%s', actual: '%s'\n",
						spkg->version, pkgfile->version);
				error = 1;
			}
		}
		if(error != 0) {
			errors++;
			*data = hbpm_list_add(*data, strdup(spkg->filename));
			free(filepath);
			continue;
		}
		free(filepath);
		/* copy over the install reason */
		pkgfile->reason = spkg->reason;
		/* copy over validation method */
		pkgfile->validation = spkg->validation;
		/* transfer oldpkg */
		pkgfile->oldpkg = spkg->oldpkg;
		spkg->oldpkg = NULL;
		i->data = pkgfile;
		/* spkg has been removed from the target list, so we can free the
		 * sync-specific fields */
		_hbpm_pkg_free_trans(spkg);
	}

	PROGRESS(handle, HBPM_PROGRESS_LOAD_START, "", 100,
			total, current);
	event.type = HBPM_EVENT_LOAD_DONE;
	EVENT(handle, &event);

	if(errors) {
		if(handle->pm_errno == HBPM_ERR_OK) {
			RET_ERR(handle, HBPM_ERR_PKG_INVALID, -1);
		}
		return -1;
	}

	return 0;
}

int _hbpm_sync_load(hbpm_handle_t *handle, hbpm_list_t **data)
{
	hbpm_list_t *i;
	size_t total = 0;
	uint64_t total_bytes = 0;
	hbpm_trans_t *trans = handle->trans;

	if(download_files(handle)) {
		return -1;
	}

#ifdef HAVE_LIBGPGME
	/* make sure all required signatures are in keyring */
	if(check_keyring(handle)) {
		return -1;
	}
#endif

	/* get the total size of all packages so we can adjust the progress bar more
	 * realistically if there are small and huge packages involved */
	for(i = trans->add; i; i = i->next) {
		hbpm_pkg_t *spkg = i->data;
		if(spkg->origin != HBPM_PKG_FROM_FILE) {
			total_bytes += spkg->size;
		}
		total++;
	}
	/* this can only happen maliciously */
	total_bytes = total_bytes ? total_bytes : 1;

	if(check_validity(handle, total, total_bytes) != 0) {
		return -1;
	}

	if(trans->flags & HBPM_TRANS_FLAG_DOWNLOADONLY) {
		return 0;
	}

	if(load_packages(handle, data, total, total_bytes)) {
		return -1;
	}

	return 0;
}

int _hbpm_sync_check(hbpm_handle_t *handle, hbpm_list_t **data)
{
	hbpm_trans_t *trans = handle->trans;
	hbpm_event_t event;

	/* fileconflict check */
	if(!(trans->flags & HBPM_TRANS_FLAG_DBONLY)) {
		event.type = HBPM_EVENT_FILECONFLICTS_START;
		EVENT(handle, &event);

		_hbpm_log(handle, HBPM_LOG_DEBUG, "looking for file conflicts\n");
		hbpm_list_t *conflict = _hbpm_db_find_fileconflicts(handle,
				trans->add, trans->remove);
		if(conflict) {
			if(data) {
				*data = conflict;
			} else {
				hbpm_list_free_inner(conflict,
						(hbpm_list_fn_free)hbpm_fileconflict_free);
				hbpm_list_free(conflict);
			}
			RET_ERR(handle, HBPM_ERR_FILE_CONFLICTS, -1);
		}

		event.type = HBPM_EVENT_FILECONFLICTS_DONE;
		EVENT(handle, &event);
	}

	/* check available disk space */
	if(handle->checkspace && !(trans->flags & HBPM_TRANS_FLAG_DBONLY)) {
		event.type = HBPM_EVENT_DISKSPACE_START;
		EVENT(handle, &event);

		_hbpm_log(handle, HBPM_LOG_DEBUG, "checking available disk space\n");
		if(_hbpm_check_diskspace(handle) == -1) {
			_hbpm_log(handle, HBPM_LOG_ERROR, _("not enough free disk space\n"));
			return -1;
		}

		event.type = HBPM_EVENT_DISKSPACE_DONE;
		EVENT(handle, &event);
	}

	return 0;
}

int _hbpm_sync_commit(hbpm_handle_t *handle)
{
	hbpm_trans_t *trans = handle->trans;

	/* remove conflicting and to-be-replaced packages */
	if(trans->remove) {
		_hbpm_log(handle, HBPM_LOG_DEBUG,
				"removing conflicting and to-be-replaced packages\n");
		/* we want the frontend to be aware of commit details */
		if(_hbpm_remove_packages(handle, 0) == -1) {
			_hbpm_log(handle, HBPM_LOG_ERROR,
					_("could not commit removal transaction\n"));
			return -1;
		}
	}

	/* install targets */
	_hbpm_log(handle, HBPM_LOG_DEBUG, "installing packages\n");
	if(_hbpm_upgrade_packages(handle) == -1) {
		_hbpm_log(handle, HBPM_LOG_ERROR, _("could not commit transaction\n"));
		return -1;
	}

	return 0;
}
