/*
 *  util.h
 *
 *  Copyright (c) 2006-2019 Pacman Development Team <pacman-dev@archlinux.org>
 *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
 *  Copyright (c) 2005 by Aurelien Foret <orelien@chez.com>
 *  Copyright (c) 2005 by Christian Hamar <krics@linuxforum.hu>
 *  Copyright (c) 2006 by David Kimpe <dnaku@frugalware.org>
 *  Copyright (c) 2005, 2006 by Miklos Vajna <vmiklos@frugalware.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef HBPM_UTIL_H
#define HBPM_UTIL_H

#include "hbpm_list.h"
#include "hbpm.h"
#include "package.h" /* hbpm_pkg_t */
#include "handle.h" /* hbpm_handle_t */
#include "util-common.h"

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stddef.h> /* size_t */
#include <sys/types.h>
#include <math.h> /* fabs */
#include <float.h> /* DBL_EPSILON */
#include <fcntl.h> /* open, close */

#include <archive.h> /* struct archive */

#ifdef ENABLE_NLS
#include <libintl.h> /* here so it doesn't need to be included elsewhere */
/* define _() as shortcut for gettext() */
#define _(str) dgettext ("libhbpm", str)
#else
#define _(s) (char *)s
#endif

void _hbpm_alloc_fail(size_t size);

#define MALLOC(p, s, action) do { p = malloc(s); if(p == NULL) { _hbpm_alloc_fail(s); action; } } while(0)
#define CALLOC(p, l, s, action) do { p = calloc(l, s); if(p == NULL) { _hbpm_alloc_fail(l * s); action; } } while(0)
#define REALLOC(p, s, action) do { void* np = realloc(p, s); if(np == NULL) { _hbpm_alloc_fail(s); action; } else { p = np; } } while(0)
/* This strdup macro is NULL safe- copying NULL will yield NULL */
#define STRDUP(r, s, action) do { if(s != NULL) { r = strdup(s); if(r == NULL) { _hbpm_alloc_fail(strlen(s)); action; } } else { r = NULL; } } while(0)
#define STRNDUP(r, s, l, action) do { if(s != NULL) { r = strndup(s, l); if(r == NULL) { _hbpm_alloc_fail(l); action; } } else { r = NULL; } } while(0)

#define FREE(p) do { free(p); p = NULL; } while(0)

#define ASSERT(cond, action) do { if(!(cond)) { action; } } while(0)

#define RET_ERR_VOID(handle, err) do { \
	_hbpm_log(handle, HBPM_LOG_DEBUG, "returning error %d from %s (%s: %d) : %s\n", err, __func__, __FILE__, __LINE__, hbpm_strerror(err)); \
	(handle)->pm_errno = (err); \
	return; } while(0)

#define RET_ERR(handle, err, ret) do { \
	_hbpm_log(handle, HBPM_LOG_DEBUG, "returning error %d from %s (%s: %d) : %s\n", err, __func__, __FILE__, __LINE__, hbpm_strerror(err)); \
	(handle)->pm_errno = (err); \
	return (ret); } while(0)

#define GOTO_ERR(handle, err, label) do { \
	_hbpm_log(handle, HBPM_LOG_DEBUG, "got error %d at %s (%s: %d) : %s\n", err, __func__, __FILE__, __LINE__, hbpm_strerror(err)); \
	(handle)->pm_errno = (err); \
	goto label; } while(0)

#define RET_ERR_ASYNC_SAFE(handle, err, ret) do { \
	(handle)->pm_errno = (err); \
	return (ret); } while(0)

#define DOUBLE_EQ(x, y) (fabs((x) - (y)) < DBL_EPSILON)

#define CHECK_HANDLE(handle, action) do { if(!(handle)) { action; } (handle)->pm_errno = HBPM_ERR_OK; } while(0)

/** Standard buffer size used throughout the library. */
#ifdef BUFSIZ
#define HBPM_BUFFER_SIZE BUFSIZ
#else
#define HBPM_BUFFER_SIZE 8192
#endif

#ifndef O_BINARY
#define O_BINARY 0
#endif

#define OPEN(fd, path, flags) do { fd = open(path, flags | O_BINARY); } while(fd == -1 && errno == EINTR)

/**
 * Used as a buffer/state holder for _hbpm_archive_fgets().
 */
struct archive_read_buffer {
	char *line;
	char *line_offset;
	size_t line_size;
	size_t max_line_size;
	size_t real_line_size;

	char *block;
	char *block_offset;
	size_t block_size;

	int ret;
};

int _hbpm_makepath(const char *path);
int _hbpm_makepath_mode(const char *path, mode_t mode);
int _hbpm_copyfile(const char *src, const char *dest);
size_t _hbpm_strip_newline(char *str, size_t len);

int _hbpm_open_archive(hbpm_handle_t *handle, const char *path,
		struct stat *buf, struct archive **archive, hbpm_errno_t error);
int _hbpm_unpack_single(hbpm_handle_t *handle, const char *archive,
		const char *prefix, const char *filename);
int _hbpm_unpack(hbpm_handle_t *handle, const char *archive, const char *prefix,
		hbpm_list_t *list, int breakfirst);

ssize_t _hbpm_files_in_directory(hbpm_handle_t *handle, const char *path, int full_count);

typedef ssize_t (*_hbpm_cb_io)(void *buf, ssize_t len, void *ctx);

int _hbpm_run_chroot(hbpm_handle_t *handle, const char *cmd, char *const argv[],
		_hbpm_cb_io in_cb, void *in_ctx);
int _hbpm_ldconfig(hbpm_handle_t *handle);
int _hbpm_str_cmp(const void *s1, const void *s2);
char *_hbpm_filecache_find(hbpm_handle_t *handle, const char *filename);
const char *_hbpm_filecache_setup(hbpm_handle_t *handle);
/* Unlike many uses of hbpm_pkgvalidation_t, _hbpm_test_checksum expects
 * an enum value rather than a bitfield. */
int _hbpm_test_checksum(const char *filepath, const char *expected, hbpm_pkgvalidation_t type);
int _hbpm_archive_fgets(struct archive *a, struct archive_read_buffer *b);
int _hbpm_splitname(const char *target, char **name, char **version,
		unsigned long *name_hash);
unsigned long _hbpm_hash_sdbm(const char *str);
off_t _hbpm_strtoofft(const char *line);
hbpm_time_t _hbpm_parsedate(const char *line);
int _hbpm_raw_cmp(const char *first, const char *second);
int _hbpm_raw_ncmp(const char *first, const char *second, size_t max);
int _hbpm_access(hbpm_handle_t *handle, const char *dir, const char *file, int amode);
int _hbpm_fnmatch_patterns(hbpm_list_t *patterns, const char *string);
int _hbpm_fnmatch(const void *pattern, const void *string);
void *_hbpm_realloc(void **data, size_t *current, const size_t required);
void *_hbpm_greedy_grow(void **data, size_t *current, const size_t required);

#ifndef HAVE_STRSEP
char *strsep(char **, const char *);
#endif

/* check exported library symbols with: nm -C -D <lib> */
#define SYMEXPORT __attribute__((visibility("default")))
#define SYMHIDDEN __attribute__((visibility("internal")))

#define UNUSED __attribute__((unused))

#endif /* HBPM_UTIL_H */
