/*
 *  testpkg.c : Test a hyperman package for validity
 *
 *  Copyright (c) 2007-2019 Pacman Development Team <pacman-dev@archlinux.org>
 *  Copyright (c) 2007 by Aaron Griffin <aaronmgriffin@gmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h> /* printf */
#include <stdarg.h> /* va_list */

#include <hbpm.h>
#include "util.h" /* For Localization */

__attribute__((format(printf, 2, 0)))
static void output_cb(hbpm_loglevel_t level, const char *fmt, va_list args)
{
	if(fmt[0] == '\0') {
		return;
	}
	switch(level) {
		case HBPM_LOG_ERROR: printf(_("error: ")); break;
		case HBPM_LOG_WARNING: printf(_("warning: ")); break;
		default: return; /* skip other messages */
	}
	vprintf(fmt, args);
}

int main(int argc, char *argv[])
{
	int retval = 1; /* default = false */
	hbpm_handle_t *handle;
	hbpm_errno_t err;
	hbpm_pkg_t *pkg = NULL;
	const int siglevel = HBPM_SIG_PACKAGE | HBPM_SIG_PACKAGE_OPTIONAL;

#if defined(ENABLE_NLS)
	bindtextdomain(PACKAGE, LOCALEDIR);
#endif

	if(argc != 2) {
		fprintf(stderr, "testpkg (hyperman) v" PACKAGE_VERSION "\n\n");
		fprintf(stderr,	_("Test a hyperman package for validity.\n\n"));
		fprintf(stderr,	_("Usage: testpkg <package file>\n"));
		return 1;
	}

	handle = hbpm_initialize(ROOTDIR, DBPATH, &err);
	if(!handle) {
		fprintf(stderr, _("cannot initialize hbpm: %s\n"), hbpm_strerror(err));
		return 1;
	}

	/* let us get log messages from libhbpm */
	hbpm_option_set_logcb(handle, output_cb);

	/* set gpgdir to default */
	hbpm_option_set_gpgdir(handle, GPGDIR);

	if(hbpm_pkg_load(handle, argv[1], 1, siglevel, &pkg) == -1
			|| pkg == NULL) {
		err = hbpm_errno(handle);
		switch(err) {
			case HBPM_ERR_PKG_NOT_FOUND:
				printf(_("Cannot find the given file.\n"));
				break;
			case HBPM_ERR_PKG_OPEN:
				printf(_("Cannot open the given file.\n"));
				break;
			case HBPM_ERR_LIBARCHIVE:
			case HBPM_ERR_PKG_INVALID:
				printf(_("Package is invalid.\n"));
				break;
			default:
				printf(_("libhbpm error: %s\n"), hbpm_strerror(err));
				break;
		}
		retval = 1;
	} else {
		hbpm_pkg_free(pkg);
		printf(_("Package is valid.\n"));
		retval = 0;
	}

	if(hbpm_release(handle) == -1) {
		fprintf(stderr, _("error releasing hbpm\n"));
	}

	return retval;
}
